#!/usr/bin/env python
# mypy: ignore-errors

import atexit
import difflib
import os
import os.path
import shlex
import sys
from collections.abc import Iterator
from subprocess import PIPE, Popen
from typing import TYPE_CHECKING

from pyglossary.core import log
from pyglossary.glossary_v2 import Glossary
from pyglossary.ui.tools.colors import (
	green,
	red,
	reset,
	yellow,
)
from pyglossary.ui.tools.format_entry import formatEntry
from pyglossary.ui.tools.word_diff import (
	formatDiff,
	xmlDiff,
)

if TYPE_CHECKING:
	from pyglossary.glossary_types import EntryType

__all__ = ["diffGlossary"]

Glossary.init()

log.setVerbosity(1)

entrySep = f"\n{'_' * 40}\n\n"

noInfo = os.getenv("GLOSSARY_DIFF_NO_INFO") == "1"


def formatInfoValueDiff(diff: "Iterator[str]") -> str:
	a = ""
	b = ""
	for part in diff:
		if part[0] == " ":
			a += part[2:]
			b += part[2:]
			continue
		if part[0] == "-":
			a += red + part[2:] + reset
			continue
		if part[0] == "+":
			b += green + part[2:] + reset
			continue
	return a + "\n" + b


def diffGlossary(  # noqa: PLR0912, PLR0913
	filename1: str,
	filename2: str,
	format1: "str | None" = None,
	format2: "str | None" = None,
	header: str = "",
	pager: bool = True,
) -> None:
	glos1 = Glossary(ui=None)
	if not glos1.directRead(filename1, format=format1):
		return

	glos2 = Glossary(ui=None)

	if not glos2.directRead(filename2, format=format2):
		return

	if pager:
		pagerCmd = ["less", "-R"]
		if os.getenv("PAGER"):
			pagerCmd = shlex.split(os.getenv("PAGER"))
		proc = Popen(
			pagerCmd,
			stdin=PIPE,
		)

		def write(msg: str):
			proc.stdin.write(msg.encode("utf-8"))

	else:
		proc = None

		def write(msg: str):
			print(msg, end="")

	if header:
		write(header + "\n")

	iter1 = iter(glos1)
	iter2 = iter(glos2)

	# infoIter1 = iter(sorted(glos1.iterInfo()))
	# infoIter2 = iter(sorted(glos2.iterInfo()))

	if noInfo:
		infoIter1 = iter([])
		infoIter2 = iter([])
	else:
		infoIter1 = glos1.iterInfo()
		infoIter2 = glos2.iterInfo()

	index1 = -1
	index2 = -1

	def nextEntry1() -> None:
		nonlocal entry1, index1
		entry1 = next(iter1)
		index1 += 1

	def nextEntry2() -> None:
		nonlocal entry2, index2
		entry2 = next(iter2)
		index2 += 1

	def printEntry(color: str, prefix: str, index: int, entry: "EntryType") -> None:
		formatted = (
			f"{color}{prefix}#{index} "
			+ formatEntry(entry).replace("\n", "\n" + color)
			+ entrySep
		)
		write(formatted)

	def printInfo(color: str, prefix: str, pair: "tuple[str, str]") -> None:
		key, value = pair
		spaces = " " * (len(prefix) + 7)
		valueColor = color + spaces + value.replace("\n", "\n" + spaces + color)
		formatted = f"{color}{prefix} Info: {key}\n{valueColor}" + entrySep
		write(formatted)

	def printChangedEntry(entry1: "EntryType", entry2: "EntryType") -> None:
		defiDiff = formatDiff(xmlDiff(entry1.defi, entry2.defi))
		entry1._defi = defiDiff
		if index1 < 0:
			ids = ""
		elif index1 == index2:
			ids = f"#{index1}"
		else:
			ids = f"A#{index1} B#{index2}"
		formatted = f"=== {yellow}{ids}{reset} " + formatEntry(entry1) + entrySep
		write(formatted)

	def printChangedInfo(key: str, value1: str, value2: str) -> str:
		valueDiff = formatInfoValueDiff(xmlDiff(value1, value2))
		printInfo(yellow, "=== ", (key, valueDiff))

	infoPair1 = None
	infoPair2 = None

	def infoStep() -> None:
		nonlocal infoPair1, infoPair2
		if infoPair1 is None:
			infoPair1 = next(infoIter1)
		if infoPair2 is None:
			infoPair2 = next(infoIter2)

		if infoPair1 == infoPair2:
			infoPair1, infoPair2 = None, None
			return

		if infoPair1[0] == infoPair2[0]:
			printChangedInfo(infoPair1[0], infoPair1[1], infoPair2[1])
			infoPair1, infoPair2 = None, None
			return

		if infoPair1[0] < infoPair2[0]:
			printInfo(red, "--- A: ", infoPair1)
			infoPair1 = None
			return

		printInfo(green, "+++ B: ", infoPair2)
		infoPair2 = None

	def printAltsChangedEntry(
		entry1: "EntryType",
		entry2: "EntryType",
		showDefi: bool = True,
	) -> None:
		ids = f"#{index1}" if index1 == index2 else f"A#{index1} B#{index2}"

		header = f"=== {yellow}{ids}{reset} "

		altsDiff = difflib.ndiff(
			[f"Alt: {alt}\n" for alt in entry1.l_word[1:]],
			[f"Alt: {alt}\n" for alt in entry2.l_word[1:]],
			linejunk=None,
			charjunk=None,
		)
		if entry1.l_word[0] == entry2.l_word[0]:
			firstWordLine = f">> {entry1.l_word[0]}"
		else:
			firstWordLine = f">> {entry1.l_word[0]} (A)\n>> {entry2.l_word[0]} (B)"
		entryFormatted = "\n".join(
			[
				firstWordLine,
				formatDiff(altsDiff),
				entry1.defi if showDefi else "",
			],
		)
		formatted = header + entryFormatted + entrySep
		write(formatted)

	count = 0
	entry1 = None
	entry2 = None

	def step() -> None:
		nonlocal count, entry1, entry2
		if entry1 is None:
			nextEntry1()
		if entry2 is None:
			nextEntry2()

		words1 = entry1.l_word
		words2 = entry2.l_word
		if words1 == words2:
			if entry1.defi == entry2.defi:
				entry1, entry2 = None, None
				return
			printChangedEntry(entry1, entry2)
			entry1, entry2 = None, None
			return

		if entry1.defi == entry2.defi and (words1[0] in words2 or words2[0] in words1):
			printAltsChangedEntry(entry1, entry2)
			entry1, entry2 = None, None
			return

		if words1 < words2:
			printEntry(red, "--- A", index1, entry1)
			entry1 = None
		else:
			printEntry(green, "+++ B", index2, entry2)
			entry2 = None

		if (count + 1) % 50 == 0:
			sys.stdout.flush()
		count += 1

	def run():  # noqa: PLR0912
		nonlocal index1, index2

		while True:
			try:
				infoStep()
			except StopIteration:
				break
			except (OSError, BrokenPipeError):
				break

		if infoPair1:
			printInfo(red, "--- A: ", infoPair1)
		if infoPair2:
			printInfo(green, "+++ B: ", infoPair2)

		for pair in infoIter1:
			printInfo(red, "--- A: ", pair)
		for pair in infoIter2:
			printInfo(green, "+++ B: ", pair)

		while True:
			try:
				step()
			except StopIteration:
				break
			except (OSError, BrokenPipeError):
				break

		if entry1:
			printEntry(red, "--- A", index1, entry1)
			index1 += 1
		if entry2:
			printEntry(green, "+++ B", index2, entry2)
			index2 += 1

		for entry in iter1:
			printEntry(red, "--- A", index1, entry)
			index1 += 1
		for entry in iter2:
			printEntry(green, "+++ B", index2, entry)
			index2 += 1

	try:
		run()
	except (OSError, BrokenPipeError):
		pass  # noqa: S110
	except Exception as e:
		print(e)
	finally:
		if proc:
			proc.communicate()
			# proc.wait()
			# proc.terminate()
			sys.stdin.flush()
			sys.stdout.flush()


# NOTE: make sure to set GIT_PAGER or config core.pager
# for example GIT_PAGER=less
# or GIT_PAGER='less -R -S -N'
def gitDiffMain() -> None:
	# print(sys.argv[1:])
	# arguments:
	# path old_file old_hex old_mode new_file new_hex new_mode
	old_hex = sys.argv[3][:7]
	new_hex = sys.argv[6][:7]

	filename1 = sys.argv[2]
	filename2 = sys.argv[1]
	header = f"{'_' * 80}\n\n### File: {filename2}  ({old_hex}..{new_hex})\n"

	resDir = filename2 + "_res"
	if os.path.isdir(resDir):
		resDirTmp = filename1 + "_res"
		os.symlink(os.path.realpath(resDir), resDirTmp)
		atexit.register(os.remove, resDirTmp)

	diffGlossary(
		filename1,
		filename2,
		format1=None,
		format2=None,
		pager=False,
		header=header,
	)


def main() -> None:
	import os

	if os.getenv("GIT_DIFF_PATH_COUNTER"):
		return gitDiffMain()

	filename1 = sys.argv[1]
	filename2 = sys.argv[2]
	format1 = None
	format2 = None
	if len(sys.argv) > 3:
		format1 = sys.argv[3]
	if len(sys.argv) > 4:
		format2 = sys.argv[4]
	filename1 = os.path.expanduser(filename1)
	filename2 = os.path.expanduser(filename2)
	diffGlossary(
		filename1,
		filename2,
		format1=format1,
		format2=format2,
		pager=True,
	)
	return None


if __name__ == "__main__":
	main()



# mypy: ignore-errors

from collections.abc import Generator, Iterator
from hashlib import sha1
from os import listdir, makedirs
from os.path import dirname, isdir, isfile, join, splitext

from pyglossary.compression import (
	compressionOpenFunc,
)
from pyglossary.core import log
from pyglossary.glossary_types import EntryType, GlossaryType
from pyglossary.option import (
	Option,
	StrOption,
)
from pyglossary.text_utils import (
	escapeNTB,
	splitByBarUnescapeNTB,
)

__all__ = [
	"Reader",
	"Writer",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

enable = True
lname = "crawler_dir"
format = "CrawlerDir"
description = "Crawler Directory"
extensions = (".crawler",)
extensionCreate = ".crawler/"
singleFile = True
kind = "directory"
wiki = ""
website = None
optionsProp: "dict[str, Option]" = {
	"compression": StrOption(
		values=["", "gz", "bz2", "lzma"],
		comment="Compression Algorithm",
	),
}


class Writer:
	_compression: str = ""

	def __init__(self, glos: GlossaryType) -> None:
		self._glos = glos
		self._filename = None

	def finish(self) -> None:
		pass

	def open(self, filename: str) -> None:
		self._filename = filename
		if not isdir(filename):
			makedirs(filename)

	@staticmethod
	def filePathFromWord(b_word: bytes) -> str:
		bw = b_word.lower()
		if len(bw) <= 2:
			return bw.hex()
		if len(bw) <= 4:
			return join(
				bw[:2].hex() + ".d",
				bw[2:].hex(),
			)
		return join(
			bw[:2].hex() + ".d",
			bw[2:4].hex() + ".d",
			bw[4:8].hex() + "-" + sha1(b_word).hexdigest()[:8],  # noqa: S324
		)

	def write(self) -> None:
		from collections import OrderedDict as odict

		from pyglossary.json_utils import dataToPrettyJson

		filename = self._filename

		wordCount = 0
		compression = self._compression
		c_open = compressionOpenFunc(compression)
		if not c_open:
			raise ValueError(f"invalid compression {compression!r}")
		while True:
			entry = yield
			if entry is None:
				break
			if entry.isData():
				continue
			fpath = join(filename, self.filePathFromWord(entry.b_word))
			if compression:
				fpath = f"{fpath}.{compression}"
			parentDir = dirname(fpath)
			if not isdir(parentDir):
				makedirs(parentDir)
			if isfile(fpath):
				log.warning(f"file exists: {fpath}")
				fpath += f"-{sha1(entry.b_defi).hexdigest()[:4]}"  # noqa: S324
			with c_open(fpath, "wt", encoding="utf-8") as _file:
				_file.write(
					f"{escapeNTB(entry.s_word)}\n{entry.defi}",
				)
			wordCount += 1

		with open(
			join(filename, "info.json"),
			mode="w",
			encoding="utf-8",
		) as infoFile:
			info = odict()
			info["name"] = self._glos.getInfo("name")
			info["wordCount"] = wordCount
			for key, value in self._glos.getExtraInfos(
				(
					"name",
					"wordCount",
				),
			).items():
				info[key] = value
			infoFile.write(dataToPrettyJson(info))


class Reader:
	def __init__(self, glos: GlossaryType) -> None:
		self._glos = glos
		self._filename = None
		self._wordCount = 0

	def open(self, filename: str) -> None:
		from pyglossary.json_utils import jsonToOrderedData

		self._filename = filename

		with open(join(filename, "info.json"), encoding="utf-8") as infoFp:
			info = jsonToOrderedData(infoFp.read())
		self._wordCount = info.pop("wordCount")
		for key, value in info.items():
			self._glos.setInfo(key, value)

	def close(self) -> None:
		pass

	def __len__(self) -> int:
		return self._wordCount

	def _fromFile(self, fpath: str) -> "EntryType":
		_, ext = splitext(fpath)
		c_open = compressionOpenFunc(ext.lstrip("."))
		if not c_open:
			log.error(f"invalid extension {ext}")
			c_open = open
		with c_open(fpath, "rt", encoding="utf-8") as _file:
			words = splitByBarUnescapeNTB(_file.readline().rstrip("\n"))
			defi = _file.read()
			return self._glos.newEntry(words, defi)

	@staticmethod
	def _listdirSortKey(name: str) -> str:
		name_nox, ext = splitext(name)
		if ext == ".d":
			return name
		return name_nox

	def _readDir(
		self,
		dpath: str,
		exclude: "set[str] | None",
	) -> "Generator[EntryType, None, None]":
		children = listdir(dpath)
		if exclude:
			children = [name for name in children if name not in exclude]
		children.sort(key=self._listdirSortKey)
		for name in children:
			cpath = join(dpath, name)
			if isfile(cpath):
				yield self._fromFile(cpath)
				continue
			if isdir(cpath):
				yield from self._readDir(cpath, None)
				continue
			log.error(f"Not a file nor a directory: {cpath}")

	def __iter__(self) -> "Iterator[EntryType]":
		yield from self._readDir(
			self._filename,
			{
				"info.json",
			},
		)



# -*- coding: utf-8 -*-

# Copyright © 2023 soshial <soshial@gmail.com> (soshial)
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

from struct import unpack
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	import io

__all__ = [
	"APPLEDICT_FILE_OFFSET",
	"guessFileOffsetLimit",
	"readInt",
	"read_2_bytes_here",
	"read_x_bytes_as_word",
]


APPLEDICT_FILE_OFFSET = 0x40
# addressing of AppleDict binary files always ignores first 0x40 bytes


def readIntPair(buffer: "io.BufferedIOBase") -> "tuple[int, int]":
	# to satisfy mymy, put them in vars with declared type
	a: int
	b: int
	a, b = unpack("ii", buffer.read(8))
	return a, b


def readInt(buffer: "io.BufferedIOBase") -> int:
	return unpack("i", buffer.read(4))[0]


def read_x_bytes_as_word(buffer: "io.BufferedIOBase", x: int) -> str:
	return buffer.read(x).decode("UTF-16LE")


def read_2_bytes_here(buffer: "io.BufferedIOBase") -> int:
	lower_byte = buffer.read(1)
	higher_byte = buffer.read(1)
	return ord(higher_byte) * 0x100 + ord(lower_byte)


def guessFileOffsetLimit(file: "io.BufferedIOBase") -> "tuple[int, int]":
	"""Returns address offset to start parsing from and EOF address."""
	file.seek(APPLEDICT_FILE_OFFSET)
	limit = readInt(file)
	intPair = readIntPair(file)

	if intPair == (0, -1):  # 0000 0000 FFFF FFFF
		return 0x20, limit
	return 0x4, limit



# -*- coding: utf-8 -*-

import gzip
import os
import re
import typing
from collections import Counter
from collections.abc import Callable, Generator, Iterator, Sequence
from os.path import (
	dirname,
	getsize,
	isdir,
	isfile,
	join,
	realpath,
	split,
	splitext,
)
from pprint import pformat
from time import perf_counter as now
from typing import (
	TYPE_CHECKING,
	Any,
	Literal,
	Protocol,
	TypeVar,
)

if TYPE_CHECKING:
	import io
	import sqlite3

	from pyglossary.langs import Lang


from pyglossary.core import log
from pyglossary.flags import ALWAYS, DEFAULT_YES
from pyglossary.glossary_types import EntryType, GlossaryType
from pyglossary.option import (
	BoolOption,
	Option,
	StrOption,
)
from pyglossary.text_utils import (
	uint32FromBytes,
	uint32ToBytes,
	uint64FromBytes,
	uint64ToBytes,
)

__all__ = [
	"Reader",
	"Writer",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]


enable = True
lname = "stardict"
format = "Stardict"
description = "StarDict (.ifo)"
extensions = (".ifo",)
extensionCreate = "-stardict/"
singleFile = False
sortOnWrite = ALWAYS
sortKeyName = "stardict"
sortEncoding = "utf-8"

kind = "directory"
wiki = "https://en.wikipedia.org/wiki/StarDict"
website = (
	"http://huzheng.org/stardict/",
	"huzheng.org/stardict",
)
# https://github.com/huzheng001/stardict-3/blob/master/dict/doc/StarDictFileFormat
optionsProp: "dict[str, Option]" = {
	"large_file": BoolOption(
		comment="Use idxoffsetbits=64 bits, for large files only",
	),
	"stardict_client": BoolOption(
		comment="Modify html entries for StarDict 3.0",
	),
	"dictzip": BoolOption(
		comment="Compress .dict file to .dict.dz",
	),
	"sametypesequence": StrOption(
		values=["", "h", "m", "x", None],
		comment="Definition format: h=html, m=plaintext, x=xdxf",
	),
	"merge_syns": BoolOption(
		comment="Write alternates to .idx instead of .syn",
	),
	"xdxf_to_html": BoolOption(
		comment="Convert XDXF entries to HTML",
	),
	"xsl": BoolOption(
		comment="Use XSL transformation",
	),
	"unicode_errors": StrOption(
		values=[
			"strict",  # raise a UnicodeDecodeError exception
			"ignore",  # just leave the character out
			"replace",  # use U+FFFD, REPLACEMENT CHARACTER
			"backslashreplace",  # insert a \xNN escape sequence
		],
		comment="What to do with Unicode decoding errors",
	),
	"audio_goldendict": BoolOption(
		comment="Convert audio links for GoldenDict (desktop)",
	),
	"audio_icon": BoolOption(
		comment="Add glossary's audio icon",
	),
	"sqlite": BoolOption(
		comment="Use SQLite to limit memory usage",
	),
}

if os.getenv("PYGLOSSARY_STARDICT_NO_FORCE_SORT") == "1":
	sortOnWrite = DEFAULT_YES

infoKeys = (
	"bookname",
	"author",
	"email",
	"website",
	"description",
	"date",
)


# re_newline = re.compile("[\n\r]+")
re_newline = re.compile("\n\r?|\r\n?")


def newlinesToSpace(text: str) -> str:
	return re_newline.sub(" ", text)


def newlinesToBr(text: str) -> str:
	return re_newline.sub("<br>", text)


def verifySameTypeSequence(s: str) -> bool:
	if not s:
		return True
	# maybe should just check it's in ("h", "m", "x")
	if not s.isalpha():
		return False
	return len(s) == 1


class XdxfTransformerType(Protocol):
	def transformByInnerString(self, text: str) -> str: ...


T_SDListItem_contra = TypeVar("T_SDListItem_contra", contravariant=True)


class T_SdList(Protocol[T_SDListItem_contra]):
	def append(self, x: T_SDListItem_contra) -> None: ...

	def __len__(self) -> int: ...

	def __iter__(self) -> "Iterator[Any]": ...

	def sort(self) -> None: ...


class MemSdList:
	def __init__(self) -> None:
		self._l: "list[Any]" = []

	def append(self, x: Any) -> None:
		self._l.append(x)

	def __len__(self) -> int:
		return len(self._l)

	def __iter__(self) -> "Iterator[Any]":
		return iter(self._l)

	def sortKey(self, item: "tuple[bytes, Any]") -> "tuple[bytes, bytes]":  # noqa: PLR6301
		return (
			item[0].lower(),
			item[0],
		)

	def sort(self) -> None:
		self._l.sort(key=self.sortKey)


class BaseSqList:
	def __init__(
		self,
		filename: str,
	) -> None:
		from sqlite3 import connect

		if isfile(filename):
			log.warning(f"Renaming {filename} to {filename}.bak")
			os.rename(filename, filename + "bak")

		self._filename = filename

		self._con: "sqlite3.Connection | None" = connect(filename)
		self._cur: "sqlite3.Cursor | None" = self._con.cursor()

		if not filename:
			raise ValueError(f"invalid {filename=}")

		self._orderBy = "word_lower, word"
		self._sorted = False
		self._len = 0

		columns = self._columns = [
			("word_lower", "TEXT"),
			("word", "TEXT"),
		] + self.getExtraColumns()

		self._columnNames = ",".join(col[0] for col in columns)

		colDefs = ",".join(f"{col[0]} {col[1]}" for col in columns)
		self._con.execute(
			f"CREATE TABLE data ({colDefs})",
		)
		self._con.execute(
			f"CREATE INDEX sortkey ON data({self._orderBy});",
		)
		self._con.commit()

	@classmethod
	def getExtraColumns(cls) -> "list[tuple[str, str]]":
		# list[(columnName, dataType)]
		return []

	def __len__(self) -> int:
		return self._len

	def append(self, item: "Sequence") -> None:
		if self._cur is None or self._con is None:
			raise RuntimeError("db is closed")
		self._len += 1
		extraN = len(self._columns) - 1
		self._cur.execute(
			f"insert into data({self._columnNames}) values (?{', ?' * extraN})",
			[item[0].lower()] + list(item),
		)
		if self._len % 1000 == 0:
			self._con.commit()

	def sort(self) -> None:
		pass

	def close(self) -> None:
		if self._cur is None or self._con is None:
			return
		self._con.commit()
		self._cur.close()
		self._con.close()
		self._con = None
		self._cur = None

	def __del__(self) -> None:
		try:
			self.close()
		except AttributeError as e:
			log.error(str(e))

	def __iter__(self) -> "Iterator[EntryType]":
		if self._cur is None:
			raise RuntimeError("db is closed")
		query = f"SELECT * FROM data ORDER BY {self._orderBy}"
		self._cur.execute(query)
		for row in self._cur:
			yield row[1:]


class IdxSqList(BaseSqList):
	@classmethod
	def getExtraColumns(cls) -> "list[tuple[str, str]]":
		# list[(columnName, dataType)]
		return [
			("idx_block", "BLOB"),
		]


class SynSqList(BaseSqList):
	@classmethod
	def getExtraColumns(cls) -> "list[tuple[str, str]]":
		# list[(columnName, dataType)]
		return [
			("entry_index", "INTEGER"),
		]


class Reader:
	_xdxf_to_html: bool = True
	_xsl: bool = False
	_unicode_errors: str = "strict"

	def __init__(self, glos: GlossaryType) -> None:
		self._glos = glos
		self.clear()

		self._xdxfTr: "XdxfTransformerType | None" = None
		self._large_file = False

		"""
		indexData format
		indexData[i] - i-th record in index file,
						a tuple (previously a list) of length 3
		indexData[i][0] - b_word (bytes)
		indexData[i][1] - definition block offset in dict file (int)
		indexData[i][2] - definition block size in dict file (int)
		REMOVED:
			indexData[i][3] - list of definitions
			indexData[i][3][j][0] - definition data
			indexData[i][3][j][1] - definition type - "h", "m" or "x"
			indexData[i][4] - list of synonyms (strings)

		synDict:
			a dict { entryIndex -> altList }
		"""

	def xdxf_setup(self) -> "XdxfTransformerType":
		if self._xsl:
			from pyglossary.xdxf.xsl_transform import XslXdxfTransformer

			return XslXdxfTransformer(encoding="utf-8")
		from pyglossary.xdxf.transform import XdxfTransformer

		return XdxfTransformer(encoding="utf-8")

	def xdxf_transform(self, text: str) -> str:
		if self._xdxfTr is None:
			self._xdxfTr = self.xdxf_setup()
		return self._xdxfTr.transformByInnerString(text)

	def close(self) -> None:
		if self._dictFile:
			self._dictFile.close()
		self.clear()

	def clear(self) -> None:
		self._dictFile: "io.IOBase | None" = None
		self._filename = ""  # base file path, no extension
		self._indexData: "list[tuple[bytes, int, int]]" = []
		self._synDict: "dict[int, list[str]]" = {}
		self._sametypesequence = ""
		self._resDir = ""
		self._resFileNames: "list[str]" = []
		self._wordCount: "int | None" = None

	def open(self, filename: str) -> None:
		if splitext(filename)[1].lower() == ".ifo":
			filename = splitext(filename)[0]
		elif isdir(filename):
			filename = join(filename, filename)
		self._filename = filename
		self._filename = realpath(self._filename)
		self.readIfoFile()
		sametypesequence = self._glos.getInfo("sametypesequence")
		if not verifySameTypeSequence(sametypesequence):
			raise LookupError(f"Invalid {sametypesequence = }")
		self._indexData = self.readIdxFile()
		self._wordCount = len(self._indexData)
		self._synDict = self.readSynFile()
		self._sametypesequence = sametypesequence
		if isfile(self._filename + ".dict.dz"):
			self._dictFile = gzip.open(self._filename + ".dict.dz", mode="rb")
		else:
			self._dictFile = open(self._filename + ".dict", mode="rb")
		self._resDir = join(dirname(self._filename), "res")
		if isdir(self._resDir):
			self._resFileNames = os.listdir(self._resDir)
		else:
			self._resDir = ""
			self._resFileNames = []
		# self.readResources()

	def __len__(self) -> int:
		if self._wordCount is None:
			raise RuntimeError(
				"StarDict: len(reader) called while reader is not open",
			)
		return self._wordCount + len(self._resFileNames)

	def readIfoFile(self) -> None:
		""".ifo file is a text file in utf-8 encoding."""
		with open(
			self._filename + ".ifo",
			mode="rb",
		) as ifoFile:
			for line in ifoFile:
				line = line.strip()  # noqa: PLW2901
				if not line:
					continue
				if line == b"StarDict's dict ifo file":
					continue
				b_key, _, b_value = line.partition(b"=")
				if not (b_key and b_value):
					continue
				try:
					key = b_key.decode("utf-8")
					value = b_value.decode("utf-8", errors=self._unicode_errors)
				except UnicodeDecodeError:
					log.error(f"ifo line is not UTF-8: {line!r}")
					continue
				self._glos.setInfo(key, value)

		idxoffsetbits = self._glos.getInfo("idxoffsetbits")
		if idxoffsetbits:
			if idxoffsetbits == "32":
				self._large_file = False
			elif idxoffsetbits == "64":
				self._large_file = True
			else:
				raise ValueError(f"invalid {idxoffsetbits = }")

	def readIdxFile(self) -> "list[tuple[bytes, int, int]]":
		if isfile(self._filename + ".idx.gz"):
			with gzip.open(self._filename + ".idx.gz") as g_file:
				idxBytes = g_file.read()
		else:
			with open(self._filename + ".idx", "rb") as _file:
				idxBytes = _file.read()

		indexData = []
		pos = 0

		if self._large_file:

			def getOffset() -> "tuple[int, int]":
				return uint64FromBytes(idxBytes[pos : pos + 8]), pos + 8
		else:

			def getOffset() -> "tuple[int, int]":
				return uint32FromBytes(idxBytes[pos : pos + 4]), pos + 4

		while pos < len(idxBytes):
			beg = pos
			pos = idxBytes.find(b"\x00", beg)
			if pos < 0:
				log.error("Index file is corrupted")
				break
			b_word = idxBytes[beg:pos]
			pos += 1
			if pos + 8 > len(idxBytes):
				log.error("Index file is corrupted")
				break
			offset, pos = getOffset()
			size = uint32FromBytes(idxBytes[pos : pos + 4])
			pos += 4
			indexData.append((b_word, offset, size))

		return indexData

	def decodeRawDefiPart(
		self,
		b_defiPart: bytes,
		i_type: int,
		unicode_errors: str,
	) -> "tuple[str, str]":
		_type = chr(i_type)

		"""
		_type: 'r'
		https://github.com/huzheng001/stardict-3/blob/master/dict/doc/StarDictFileFormat#L431
		Resource file list.
		The content can be:
		img:pic/example.jpg	// Image file
		snd:apple.wav		// Sound file
		vdo:film.avi		// Video file
		att:file.bin		// Attachment file
		More than one line is supported as a list of available files.
		StarDict will find the files in the Resource Storage.
		The image will be shown, the sound file will have a play button.
		You can "save as" the attachment file and so on.
		The file list must be a utf-8 string ending with '\0'.
		Use '\n' for separating new lines.
		Use '/' character as directory separator.
		"""

		_format = {
			"m": "m",
			"t": "m",
			"y": "m",
			"g": "h",
			"h": "h",
			"x": "x",
		}.get(_type, "")

		if not _format:
			log.warning(f"Definition type {_type!r} is not supported")

		_defi = b_defiPart.decode("utf-8", errors=unicode_errors)

		# log.info(f"{_type}->{_format}: {_defi}".replace("\n", "")[:120])

		if _format == "x" and self._xdxf_to_html:
			_defi = self.xdxf_transform(_defi)
			_format = "h"

		return _format, _defi

	def renderRawDefiList(
		self,
		rawDefiList: "list[tuple[bytes, int]]",
		unicode_errors: str,
	) -> "tuple[str, str]":
		if len(rawDefiList) == 1:
			b_defiPart, i_type = rawDefiList[0]
			_format, _defi = self.decodeRawDefiPart(
				b_defiPart=b_defiPart,
				i_type=i_type,
				unicode_errors=unicode_errors,
			)
			return _defi, _format

		defiFormatSet = set()
		defisWithFormat = []
		for b_defiPart, i_type in rawDefiList:
			_format, _defi = self.decodeRawDefiPart(
				b_defiPart=b_defiPart,
				i_type=i_type,
				unicode_errors=unicode_errors,
			)
			defisWithFormat.append((_defi, _format))
			defiFormatSet.add(_format)

		if len(defiFormatSet) == 1:
			defis = [_defi for _defi, _ in defisWithFormat]
			_format = defiFormatSet.pop()
			if _format == "h":
				return "\n<hr>".join(defis), _format
			return "\n".join(defis), _format

		if not defiFormatSet:
			log.error(f"empty defiFormatSet, {rawDefiList=}")
			return "", ""

		# convert plaintext or xdxf to html
		defis = []
		for _defi, _format in defisWithFormat:
			if _format == "m":
				_defi = _defi.replace("\n", "<br/>")
				_defi = f"<pre>{_defi}</pre>"
			elif _format == "x":
				_defi = self.xdxf_transform(_defi)
			defis.append(_defi)
		return "\n<hr>\n".join(defis), "h"

	def __iter__(self) -> "Iterator[EntryType]":  # noqa: PLR0912
		indexData = self._indexData
		synDict = self._synDict
		sametypesequence = self._sametypesequence
		dictFile = self._dictFile
		unicode_errors = self._unicode_errors

		if not dictFile:
			raise RuntimeError("iterating over a reader while it's not open")

		if not indexData:
			log.warning("indexData is empty")
			return

		for entryIndex, (b_word, defiOffset, defiSize) in enumerate(indexData):
			if not b_word:
				continue

			dictFile.seek(defiOffset)
			if dictFile.tell() != defiOffset:
				log.error(f"Unable to read definition for word {b_word!r}")
				continue

			b_defiBlock = dictFile.read(defiSize)

			if len(b_defiBlock) != defiSize:
				log.error(f"Unable to read definition for word {b_word!r}")
				continue

			if sametypesequence:
				rawDefiList = self.parseDefiBlockCompact(
					b_defiBlock,
					sametypesequence,
				)
			else:
				rawDefiList = self.parseDefiBlockGeneral(b_defiBlock)

			if rawDefiList is None:
				log.error(f"Data file is corrupted. Word {b_word!r}")
				continue

			word: "str | list[str]"
			word = b_word.decode("utf-8", errors=unicode_errors)
			try:
				alts = synDict[entryIndex]
			except KeyError:  # synDict is dict
				pass
			else:
				word = [word] + alts

			defi, defiFormat = self.renderRawDefiList(
				rawDefiList,
				unicode_errors,
			)

			# FIXME:
			# defi = defi.replace(' src="./res/', ' src="./')
			yield self._glos.newEntry(word, defi, defiFormat=defiFormat)

		if isdir(self._resDir):
			for fname in os.listdir(self._resDir):
				fpath = join(self._resDir, fname)
				with open(fpath, "rb") as _file:
					yield self._glos.newDataEntry(
						fname,
						_file.read(),
					)

	def readSynFile(self) -> "dict[int, list[str]]":
		"""Return synDict, a dict { entryIndex -> altList }."""
		if self._wordCount is None:
			raise RuntimeError("self._wordCount is None")

		unicode_errors = self._unicode_errors

		synBytes = b""
		if isfile(self._filename + ".syn"):
			with open(self._filename + ".syn", mode="rb") as _file:
				synBytes = _file.read()
		elif isfile(self._filename + ".syn.dz"):
			with gzip.open(self._filename + ".syn.dz", mode="rb") as _zfile:
				synBytes = _zfile.read()
		else:
			return {}

		synBytesLen = len(synBytes)
		synDict: "dict[int, list[str]]" = {}
		pos = 0
		while pos < synBytesLen:
			beg = pos
			pos = synBytes.find(b"\x00", beg)
			if pos < 0:
				log.error("Synonym file is corrupted")
				break
			b_alt = synBytes[beg:pos]  # b_alt is bytes
			pos += 1
			if pos + 4 > len(synBytes):
				log.error("Synonym file is corrupted")
				break
			entryIndex = uint32FromBytes(synBytes[pos : pos + 4])
			pos += 4
			if entryIndex >= self._wordCount:
				log.error(
					"Corrupted synonym file. "
					f"Word {b_alt!r} references invalid item",
				)
				continue

			s_alt = b_alt.decode("utf-8", errors=unicode_errors)
			# s_alt is str
			try:
				synDict[entryIndex].append(s_alt)
			except KeyError:
				synDict[entryIndex] = [s_alt]

		return synDict

	@staticmethod
	def parseDefiBlockCompact(
		b_block: bytes,
		sametypesequence: str,
	) -> "list[tuple[bytes, int]] | None":
		"""
		Parse definition block when sametypesequence option is specified.

		Return a list of (b_defi, defiFormatCode) tuples
			where b_defi is a bytes instance
			and defiFormatCode is int, so: defiFormat = chr(defiFormatCode)
		"""
		b_sametypesequence = sametypesequence.encode("utf-8")
		if not b_sametypesequence:
			raise ValueError(f"{b_sametypesequence = }")
		res = []
		i = 0
		for t in b_sametypesequence[:-1]:
			if i >= len(b_block):
				return None
			if bytes([t]).islower():
				beg = i
				i = b_block.find(b"\x00", beg)
				if i < 0:
					return None
				res.append((b_block[beg:i], t))
				i += 1
			else:
				# assert bytes([t]).isupper()
				if i + 4 > len(b_block):
					return None
				size = uint32FromBytes(b_block[i : i + 4])
				i += 4
				if i + size > len(b_block):
					return None
				res.append((b_block[i : i + size], t))
				i += size

		if i >= len(b_block):
			return None
		t = b_sametypesequence[-1]
		if bytes([t]).islower():
			if 0 in b_block[i:]:
				return None
			res.append((b_block[i:], t))
		else:
			# assert bytes([t]).isupper()
			res.append((b_block[i:], t))

		return res

	@staticmethod
	def parseDefiBlockGeneral(
		b_block: bytes,
	) -> "list[tuple[bytes, int]] | None":
		"""
		Parse definition block when sametypesequence option is not specified.

		Return a list of (b_defi, defiFormatCode) tuples
			where b_defi is a bytes instance
			and defiFormatCode is int, so: defiFormat = chr(defiFormatCode)
		"""
		res = []
		i = 0
		while i < len(b_block):
			t = b_block[i]
			if not bytes([t]).isalpha():
				return None
			i += 1
			if bytes([t]).islower():
				beg = i
				i = b_block.find(b"\x00", beg)
				if i < 0:
					return None
				res.append((b_block[beg:i], t))
				i += 1
			else:
				# assert bytes([t]).isupper()
				if i + 4 > len(b_block):
					return None
				size = uint32FromBytes(b_block[i : i + 4])
				i += 4
				if i + size > len(b_block):
					return None
				res.append((b_block[i : i + size], t))
				i += size
		return res

	# def readResources(self):
	# 	if not isdir(self._resDir):
	# 		resInfoPath = join(baseDirPath, "res.rifo")
	# 		if isfile(resInfoPath):
	# 			log.warning(
	# 				"StarDict resource database is not supported. Skipping"
	# 			)


class Writer:
	_large_file: bool = False
	_dictzip: bool = True
	_sametypesequence: "Literal['', 'h', 'm', 'x'] | None" = ""
	_stardict_client: bool = False
	_merge_syns: bool = False
	_audio_goldendict: bool = False
	_audio_icon: bool = True
	_sqlite: bool = False

	def __init__(self, glos: GlossaryType) -> None:
		self._glos = glos
		self._filename = ""
		self._resDir = ""
		self._sourceLang: "Lang | None" = None
		self._targetLang: "Lang | None" = None
		self._p_pattern = re.compile(
			"<p( [^<>]*?)?>(.*?)</p>",
			re.DOTALL,
		)
		self._br_pattern = re.compile(
			"<br[ /]*>",
			re.IGNORECASE,
		)
		self._re_audio_link = re.compile(
			'<a (type="sound" )?([^<>]*? )?href="sound://([^<>"]+)"( .*?)?>(.*?)</a>',
		)

	def finish(self) -> None:
		self._filename = ""
		self._resDir = ""
		self._sourceLang = None
		self._targetLang = None

	def open(self, filename: str) -> None:
		log.debug(f"open: {filename = }")
		fileBasePath = filename
		##
		if splitext(filename)[1].lower() == ".ifo":
			fileBasePath = splitext(filename)[0]
		elif filename.endswith(os.sep):
			if not isdir(filename):
				os.makedirs(filename)
			fileBasePath = join(filename, split(filename[:-1])[-1])
		elif isdir(filename):
			fileBasePath = join(filename, split(filename)[-1])

		parentDir = split(fileBasePath)[0]
		if not isdir(parentDir):
			log.info(f"Creating directory {parentDir}")
			os.mkdir(parentDir)
		##
		if fileBasePath:
			fileBasePath = realpath(fileBasePath)
		self._filename = fileBasePath
		self._resDir = join(dirname(fileBasePath), "res")
		self._sourceLang = self._glos.sourceLang
		self._targetLang = self._glos.targetLang
		if self._sametypesequence:
			log.debug(f"Using write option sametypesequence={self._sametypesequence}")
		elif self._sametypesequence is not None:
			stat = self._glos.collectDefiFormat(100)
			log.debug(f"defiFormat stat: {stat}")
			if stat:
				if stat["m"] > 0.97:
					log.info("Auto-selecting sametypesequence=m")
					self._sametypesequence = "m"
				elif stat["h"] > 0.5:
					log.info("Auto-selecting sametypesequence=h")
					self._sametypesequence = "h"

	def write(self) -> "Generator[None, EntryType, None]":
		from pyglossary.os_utils import runDictzip

		if self._sametypesequence:
			if self._merge_syns:
				yield from self.writeCompactMergeSyns(self._sametypesequence)
			else:
				yield from self.writeCompact(self._sametypesequence)
		elif self._merge_syns:
			yield from self.writeGeneralMergeSyns()
		else:
			yield from self.writeGeneral()
		if self._dictzip:
			runDictzip(f"{self._filename}.dict")
			syn_file = f"{self._filename}.syn"
			if not self._merge_syns and os.path.exists(syn_file):
				runDictzip(syn_file)

	def fixDefi(self, defi: str, defiFormat: str) -> str:
		# for StarDict 3.0:
		if self._stardict_client and defiFormat == "h":
			defi = self._p_pattern.sub("\\2<br>", defi)
			# if there is </p> left without opening, replace with <br>
			defi = defi.replace("</p>", "<br>")
			defi = self._br_pattern.sub("<br>", defi)

		if self._audio_goldendict:
			if self._audio_icon:
				defi = self._re_audio_link.sub(
					r'<audio src="\3">\5</audio>',
					defi,
				)
			else:
				defi = self._re_audio_link.sub(
					r'<audio src="\3"></audio>',
					defi,
				)

		# FIXME:
		# defi = defi.replace(' src="./', ' src="./res/')
		return defi

	def newIdxList(self) -> "T_SdList":
		if not self._sqlite:
			return MemSdList()
		return IdxSqList(join(self._glos.tmpDataDir, "stardict-idx.db"))

	def newSynList(self) -> "T_SdList":
		if not self._sqlite:
			return MemSdList()
		return SynSqList(join(self._glos.tmpDataDir, "stardict-syn.db"))

	def dictMarkToBytesFunc(self) -> "tuple[Callable, int]":
		if self._large_file:
			return uint64ToBytes, 0xFFFFFFFFFFFFFFFF

		return uint32ToBytes, 0xFFFFFFFF

	def writeCompact(self, defiFormat: str) -> "Generator[None, EntryType, None]":
		"""
		Build StarDict dictionary with sametypesequence option specified.
		Every item definition consists of a single article.
		All articles have the same format, specified in defiFormat parameter.

		defiFormat: format of article definition: h - html, m - plain text
		"""
		log.debug(f"writeCompact: {defiFormat=}")
		dictMark = 0
		altIndexList = self.newSynList()

		dictFile = open(self._filename + ".dict", "wb")
		idxFile = open(self._filename + ".idx", "wb")

		dictMarkToBytes, dictMarkMax = self.dictMarkToBytesFunc()

		t0 = now()
		wordCount = 0
		if not isdir(self._resDir):
			os.mkdir(self._resDir)

		entryIndex = -1
		while True:
			entry = yield
			if entry is None:
				break
			if entry.isData():
				entry.save(self._resDir)
				continue
			entryIndex += 1

			words = entry.l_word  # list of strs
			word = words[0]  # str
			defi = self.fixDefi(entry.defi, defiFormat)
			# defi is str

			for alt in words[1:]:
				altIndexList.append((alt.encode("utf-8"), entryIndex))

			b_dictBlock = defi.encode("utf-8")
			dictFile.write(b_dictBlock)
			blockLen = len(b_dictBlock)

			b_idxBlock = (
				word.encode("utf-8")
				+ b"\x00"
				+ dictMarkToBytes(dictMark)
				+ uint32ToBytes(blockLen)
			)
			idxFile.write(b_idxBlock)

			dictMark += blockLen
			wordCount += 1

			if dictMark > dictMarkMax:
				log.error(
					f"StarDict: {dictMark = } is too big, "
					"set option large_file=true",
				)
				break

		dictFile.close()
		idxFile.close()
		if not os.listdir(self._resDir):
			os.rmdir(self._resDir)
		log.info(f"Writing dict file took {now() - t0:.2f} seconds")

		self.writeSynFile(altIndexList)
		self.writeIfoFile(
			wordCount,
			len(altIndexList),
			defiFormat=defiFormat,
		)

	def writeGeneral(self) -> "Generator[None, EntryType, None]":
		"""
		Build StarDict dictionary in general case.
		Every item definition may consist of an arbitrary number of articles.
		sametypesequence option is not used.
		"""
		log.debug("writeGeneral")
		dictMark = 0
		altIndexList = self.newSynList()

		dictFile = open(self._filename + ".dict", "wb")
		idxFile = open(self._filename + ".idx", "wb")

		t0 = now()
		wordCount = 0
		defiFormatCounter: "typing.Counter[str]" = Counter()
		if not isdir(self._resDir):
			os.mkdir(self._resDir)

		dictMarkToBytes, dictMarkMax = self.dictMarkToBytesFunc()

		entryIndex = -1
		while True:
			entry = yield
			if entry is None:
				break
			if entry.isData():
				entry.save(self._resDir)
				continue
			entryIndex += 1

			entry.detectDefiFormat()  # call no more than once
			defiFormat = entry.defiFormat
			defiFormatCounter[defiFormat] += 1
			if defiFormat not in {"h", "m", "x"}:
				log.error(f"invalid {defiFormat=}, using 'm'")
				defiFormat = "m"

			words = entry.l_word  # list of strs
			word = words[0]  # str
			defi = self.fixDefi(entry.defi, defiFormat)
			# defi is str

			for alt in words[1:]:
				altIndexList.append((alt.encode("utf-8"), entryIndex))

			b_dictBlock = (defiFormat + defi).encode("utf-8") + b"\x00"
			dictFile.write(b_dictBlock)
			blockLen = len(b_dictBlock)

			b_idxBlock = (
				word.encode("utf-8")
				+ b"\x00"
				+ dictMarkToBytes(dictMark)
				+ uint32ToBytes(blockLen)
			)
			idxFile.write(b_idxBlock)

			dictMark += blockLen
			wordCount += 1

			if dictMark > dictMarkMax:
				log.error(
					f"StarDict: {dictMark = } is too big, "
					"set option large_file=true",
				)
				break

		dictFile.close()
		idxFile.close()
		if not os.listdir(self._resDir):
			os.rmdir(self._resDir)
		log.info(f"Writing dict file took {now() - t0:.2f} seconds")
		log.debug("defiFormatsCount = " + pformat(defiFormatCounter.most_common()))

		self.writeSynFile(altIndexList)
		self.writeIfoFile(
			wordCount,
			len(altIndexList),
			defiFormat="",
		)

	def writeSynFile(self, altIndexList: "T_SdList[tuple[bytes, int]]") -> None:
		"""Build .syn file."""
		if not altIndexList:
			return

		log.info(f"Sorting {len(altIndexList)} synonyms...")
		t0 = now()

		altIndexList.sort()
		# 28 seconds with old sort key (converted from custom cmp)
		# 0.63 seconds with my new sort key
		# 0.20 seconds without key function (default sort)

		log.info(
			f"Sorting {len(altIndexList)} synonyms took {now() - t0:.2f} seconds",
		)
		log.info(f"Writing {len(altIndexList)} synonyms...")
		t0 = now()
		with open(self._filename + ".syn", "wb") as synFile:
			synFile.write(
				b"".join(
					b_alt + b"\x00" + uint32ToBytes(entryIndex)
					for b_alt, entryIndex in altIndexList
				),
			)
		log.info(
			f"Writing {len(altIndexList)} synonyms took {now() - t0:.2f} seconds",
		)

	def writeCompactMergeSyns(
		self,
		defiFormat: str,
	) -> "Generator[None, EntryType, None]":
		"""
		Build StarDict dictionary with sametypesequence option specified.
		Every item definition consists of a single article.
		All articles have the same format, specified in defiFormat parameter.

		defiFormat - format of article definition: h - html, m - plain text
		"""
		log.debug(f"writeCompactMergeSyns: {defiFormat=}")
		dictMark = 0

		idxBlockList = self.newIdxList()
		altIndexList = self.newSynList()

		dictFile = open(self._filename + ".dict", "wb")

		t0 = now()
		if not isdir(self._resDir):
			os.mkdir(self._resDir)

		dictMarkToBytes, dictMarkMax = self.dictMarkToBytesFunc()

		entryIndex = -1
		while True:
			entry = yield
			if entry is None:
				break
			if entry.isData():
				entry.save(self._resDir)
				continue
			entryIndex += 1

			words = entry.l_word  # list of strs
			word = words[0]  # str
			defi = self.fixDefi(entry.defi, defiFormat)
			# defi is str

			b_dictBlock = defi.encode("utf-8")
			dictFile.write(b_dictBlock)
			blockLen = len(b_dictBlock)

			blockData = dictMarkToBytes(dictMark) + uint32ToBytes(blockLen)
			for word in words:
				idxBlockList.append((word.encode("utf-8"), blockData))

			dictMark += blockLen

			if dictMark > dictMarkMax:
				log.error(
					f"StarDict: {dictMark = } is too big, "
					"set option large_file=true",
				)
				break

		wordCount = self.writeIdxFile(idxBlockList)

		dictFile.close()
		if not os.listdir(self._resDir):
			os.rmdir(self._resDir)
		log.info(f"Writing dict file took {now() - t0:.2f} seconds")

		self.writeIfoFile(
			wordCount,
			len(altIndexList),
			defiFormat=defiFormat,
		)

	def writeGeneralMergeSyns(self) -> "Generator[None, EntryType, None]":
		"""
		Build StarDict dictionary in general case.
		Every item definition may consist of an arbitrary number of articles.
		sametypesequence option is not used.
		"""
		log.debug("writeGeneralMergeSyns")
		dictMark = 0
		idxBlockList = self.newIdxList()
		altIndexList = self.newSynList()

		dictFile = open(self._filename + ".dict", "wb")

		t0 = now()
		wordCount = 0
		defiFormatCounter: "typing.Counter[str]" = Counter()
		if not isdir(self._resDir):
			os.mkdir(self._resDir)

		dictMarkToBytes, dictMarkMax = self.dictMarkToBytesFunc()

		entryIndex = -1
		while True:
			entry = yield
			if entry is None:
				break
			if entry.isData():
				entry.save(self._resDir)
				continue
			entryIndex += 1

			entry.detectDefiFormat()  # call no more than once
			defiFormat = entry.defiFormat
			defiFormatCounter[defiFormat] += 1
			if defiFormat not in {"h", "m", "x"}:
				log.error(f"invalid {defiFormat=}, using 'm'")
				defiFormat = "m"

			words = entry.l_word  # list of strs
			word = words[0]  # str
			defi = self.fixDefi(entry.defi, defiFormat)
			# defi is str

			b_dictBlock = (defiFormat + defi).encode("utf-8") + b"\x00"
			dictFile.write(b_dictBlock)
			blockLen = len(b_dictBlock)

			blockData = dictMarkToBytes(dictMark) + uint32ToBytes(blockLen)
			for word in words:
				idxBlockList.append((word.encode("utf-8"), blockData))

			dictMark += blockLen

			if dictMark > dictMarkMax:
				log.error(
					f"StarDict: {dictMark = } is too big, "
					"set option large_file=true",
				)
				break

		wordCount = self.writeIdxFile(idxBlockList)

		dictFile.close()
		if not os.listdir(self._resDir):
			os.rmdir(self._resDir)
		log.info(f"Writing dict file took {now() - t0:.2f} seconds")
		log.debug("defiFormatsCount = " + pformat(defiFormatCounter.most_common()))

		self.writeIfoFile(
			wordCount,
			len(altIndexList),
			defiFormat="",
		)

	def writeIdxFile(self, indexList: "T_SdList[tuple[bytes, bytes]]") -> int:
		filename = self._filename + ".idx"
		if not indexList:
			return 0

		log.info(f"Sorting {len(indexList)} items...")
		t0 = now()

		indexList.sort()
		log.info(
			f"Sorting {len(indexList)} {filename} took {now() - t0:.2f} seconds",
		)
		log.info(f"Writing {len(indexList)} index entries...")
		t0 = now()
		with open(filename, mode="wb") as indexFile:
			indexFile.write(b"".join(key + b"\x00" + value for key, value in indexList))
		log.info(
			f"Writing {len(indexList)} {filename} took {now() - t0:.2f} seconds",
		)
		return len(indexList)

	def writeIfoFile(
		self,
		wordCount: int,
		synWordCount: int,
		defiFormat: str = "",
	) -> None:
		"""Build .ifo file."""
		glos = self._glos
		bookname = newlinesToSpace(glos.getInfo("name"))
		indexFileSize = getsize(self._filename + ".idx")

		sourceLang = self._sourceLang
		targetLang = self._targetLang
		if sourceLang and targetLang:
			langs = f"{sourceLang.code}-{targetLang.code}"
			if langs not in bookname.lower():
				bookname = f"{bookname} ({langs})"
			log.info(f"bookname: {bookname}")

		ifo: "list[tuple[str, str]]" = [
			("version", "3.0.0"),
			("bookname", bookname),
			("wordcount", str(wordCount)),
			("idxfilesize", str(indexFileSize)),
		]
		if self._large_file:
			ifo.append(("idxoffsetbits", "64"))
		if defiFormat:
			ifo.append(("sametypesequence", defiFormat))
		if synWordCount > 0:
			ifo.append(("synwordcount", str(synWordCount)))

		desc = glos.getInfo("description")
		_copyright = glos.getInfo("copyright")
		if _copyright:
			desc = f"{_copyright}\n{desc}"
		publisher = glos.getInfo("publisher")
		if publisher:
			desc = f"Publisher: {publisher}\n{desc}"

		for key in infoKeys:
			if key in {
				"bookname",
				"description",
			}:
				continue
			value = glos.getInfo(key)
			if not value:
				continue
			value = newlinesToSpace(value)
			ifo.append((key, value))

		ifo.append(("description", newlinesToBr(desc)))

		ifoStr = "StarDict's dict ifo file\n"
		for key, value in ifo:
			ifoStr += f"{key}={value}\n"
		with open(
			self._filename + ".ifo",
			mode="w",
			encoding="utf-8",
			newline="\n",
		) as ifoFile:
			ifoFile.write(ifoStr)



# -*- coding: utf-8 -*-
# Read Octopus MDict dictionary format, mdx(dictionary)/mdd(data)
#
# Copyright © 2013 Xiaoqiang Wang <xiaoqiangwang AT gmail DOT com>
# Copyright © 2013-2021 Saeed Rasooli <saeed.gnu@gmail.com>
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

import gc
import os
import re
import sys
from collections.abc import Iterator
from os.path import dirname, extsep, isfile, join, splitext
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from pyglossary.glossary_types import EntryType, GlossaryType
	from pyglossary.plugin_lib.readmdict import MDD, MDX


from pyglossary.core import log
from pyglossary.option import (
	BoolOption,
	EncodingOption,
	Option,
)
from pyglossary.text_utils import toStr

__all__ = [
	"Reader",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

enable = True
lname = "octopus_mdict"
format = "OctopusMdict"
description = "Octopus MDict (.mdx)"
extensions = (".mdx",)
extensionCreate = ""
singleFile = False
kind = "binary"
wiki = ""
website = (
	"https://www.mdict.cn/wp/?page_id=5325&lang=en",
	"Download | MDict.cn",
)
optionsProp: "dict[str, Option]" = {
	"encoding": EncodingOption(),
	"substyle": BoolOption(
		comment="Enable substyle",
	),
	"same_dir_data_files": BoolOption(
		comment="Read data files from same directory",
	),
	"audio": BoolOption(
		comment="Enable audio objects",
	),
}

extraDocs = [
	(
		"`python-lzo` is required for **some** MDX glossaries.",
		"""First try converting your MDX file, if failed (`AssertionError` probably),
then try to install [LZO library and Python binding](./doc/lzo.md).""",
	),
]


class Reader:
	_encoding: str = ""
	_substyle: bool = True
	_same_dir_data_files: bool = False
	_audio: bool = False

	def __init__(self, glos: "GlossaryType") -> None:
		self._glos = glos
		self.clear()
		self._re_internal_link = re.compile("href=([\"'])(entry://|[dx]:)")
		self._re_audio_link = re.compile(
			'<a (type="sound" )?([^<>]*? )?href="sound://([^<>"]+)"( .*?)?>(.*?)</a>',
		)

	def clear(self) -> None:
		self._filename = ""
		self._mdx: "MDX | None" = None
		self._mdd: "list[MDD]" = []
		self._wordCount = 0
		self._dataEntryCount = 0

		# dict of mainWord -> newline-separated alternatives
		self._linksDict: "dict[str, str]" = {}

	def open(self, filename: str) -> None:
		from pyglossary.plugin_lib.readmdict import MDD, MDX

		self._filename = filename
		self._mdx = MDX(filename, self._encoding, self._substyle)

		"""
			multiple MDD files are supported with this naming schema:
				FILE.mdx
				FILE.mdd
				FILE.1.mdd
				FILE.2.mdd
				FILE.3.mdd
		"""

		filenameNoExt, _ext = splitext(self._filename)
		mddBase = filenameNoExt + extsep
		for fname in (f"{mddBase}mdd", f"{mddBase}1.mdd"):
			if isfile(fname):
				self._mdd.append(MDD(fname))
		mddN = 2
		while isfile(f"{mddBase}{mddN}.mdd"):
			self._mdd.append(MDD(f"{mddBase}{mddN}.mdd"))
			mddN += 1

		dataEntryCount = 0
		for mdd in self._mdd:
			dataEntryCount += len(mdd)
		self._dataEntryCount = dataEntryCount
		log.info(f"Found {len(self._mdd)} mdd files with {dataEntryCount} entries")

		# from pprint import pformat
		# log.debug("mdx.header = " + pformat(self._mdx.header))
		# for key, value in self._mdx.header.items():
		# 	key = key.lower()
		# 	self._glos.setInfo(key, value)
		try:
			title = toStr(self._mdx.header[b"Title"])
		except KeyError:
			pass
		else:
			title = title.strip()
			if title:
				self._glos.setInfo("name", title)
		desc = toStr(self._mdx.header.get(b"Description", ""))
		if desc:
			self._glos.setInfo("description", desc)

		self.loadLinks()

	def loadLinks(self) -> None:
		from pyglossary.plugin_lib.readmdict import MDX

		mdx = self._mdx
		if mdx is None:
			raise ValueError("mdx is None")

		log.info("extracting links...")
		linksDict: "dict[str, str]" = {}
		word = ""
		wordCount = 0
		for b_word, b_defi in mdx.items():
			word = b_word.decode("utf-8")
			defi = b_defi.decode("utf-8").strip()
			if defi.startswith("@@@LINK="):
				if not word:
					log.warning(f"unexpected defi: {defi}")
					continue
				mainWord = defi[8:]
				if mainWord in linksDict:
					linksDict[mainWord] += "\n" + word
				else:
					linksDict[mainWord] = word
				continue
			wordCount += 1

		log.info(
			f"extracting links done, sizeof(linksDict)={sys.getsizeof(linksDict)}",
		)
		log.info(f"{wordCount = }")
		self._linksDict = linksDict
		self._wordCount = wordCount
		self._mdx = MDX(self._filename, self._encoding, self._substyle)

	def fixDefi(self, defi: str) -> str:
		defi = self._re_internal_link.sub(r"href=\1bword://", defi)
		defi = defi.replace(' src="file://', ' src=".')

		if self._audio:
			# \5 is the possible elements between <a ...> and </a>
			# but anything between <audio...> and </audio> is completely
			# ignored by Aaard2 Web and browser
			# and there is no point adding it after </audio>
			# which makes it shown after audio controls

			# GoldenDict acts completely different, so must use
			# audio_goldendict=True option in StarDict writer instead.

			defi = self._re_audio_link.sub(
				r'<audio controls src="\3"></audio>',
				defi,
			)

		return defi

	def __iter__(self) -> "Iterator[EntryType]":
		if self._mdx is None:
			log.error("trying to iterate on a closed MDX file")
			return

		glos = self._glos
		linksDict = self._linksDict
		for b_word, b_defi in self._mdx.items():
			word = b_word.decode("utf-8")
			defi = b_defi.decode("utf-8").strip()
			if defi.startswith("@@@LINK="):
				continue
			defi = self.fixDefi(defi)
			words = word
			altsStr = linksDict.get(word, "")
			if altsStr:
				words = [word] + altsStr.split("\n")
			yield glos.newEntry(words, defi)

		self._mdx = None
		del linksDict
		self._linksDict = {}
		gc.collect()

		if self._same_dir_data_files:
			dirPath = dirname(self._filename)
			for fname in os.listdir(dirPath):
				ext = splitext(fname)[1].lower()
				if ext in {".mdx", ".mdd"}:
					continue
				fpath = join(dirPath, fname)
				if not isfile(fpath):
					continue
				with open(fpath, mode="rb") as _file:
					b_data = _file.read()
				yield glos.newDataEntry(fname, b_data)

		for mdd in self._mdd:
			try:
				for b_fname, b_data in mdd.items():
					fname = toStr(b_fname)
					fname = fname.replace("\\", os.sep).lstrip(os.sep)
					yield glos.newDataEntry(fname, b_data)
			except Exception:  # noqa: PERF203
				log.exception(f"Error reading {mdd.filename}")
		self._mdd = []

	def __len__(self) -> int:
		return self._wordCount + self._dataEntryCount

	def close(self) -> None:
		self.clear()



# -*- coding: utf-8 -*-
# glossary.py
#
# Copyright © 2008-2022 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# This file is part of PyGlossary project, https://github.com/ilius/pyglossary
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

import os
import os.path
from collections import OrderedDict as odict
from contextlib import suppress
from dataclasses import dataclass
from os.path import (
	isdir,
	isfile,
	join,
	relpath,
)
from pickle import dumps as pickle_dumps
from pickle import loads as pickle_loads
from time import perf_counter as now
from typing import (
	TYPE_CHECKING,
	cast,
)
from zlib import compress as zlib_compress
from zlib import decompress as zlib_decompress

from . import core
from .core import (
	cacheDir,
	log,
)
from .entry import DataEntry, Entry
from .entry_filters import (
	EntryFilterType,
	PreventDuplicateWords,
	RemoveHtmlTagsAll,
	ShowMaxMemoryUsage,
	ShowProgressBar,
	StripFullHtml,
	entryFiltersRules,
)
from .entry_list import EntryList
from .flags import (
	ALWAYS,
	DEFAULT_YES,
	NEVER,
)
from .glossary_info import GlossaryInfo
from .glossary_progress import GlossaryProgress
from .glossary_types import (
	EntryListType,
	EntryType,
	GlossaryExtendedType,
	GlossaryType,
	RawEntryType,
)
from .glossary_utils import Error, ReadError, WriteError, splitFilenameExt
from .info import c_name
from .os_utils import rmtree, showMemoryUsage
from .plugin_manager import PluginManager
from .sort_keys import defaultSortKeyName, lookupSortKey

if TYPE_CHECKING:
	from collections.abc import Callable, Iterator
	from typing import (
		Any,
	)

	from .entry_base import MultiStr
	from .plugin_prop import PluginProp
	from .sort_keys import NamedSortKey
	from .ui_type import UIType


__all__ = ["ConvertArgs", "Glossary", "GlossaryCommon"]

# sortKeyType = Callable[
# 	[[list[str]],
# 	Any,
# ]


@dataclass(slots=True, frozen=True)
class ConvertArgs:
	inputFilename: str
	inputFormat: str = ""
	direct: "bool | None" = None
	outputFilename: str = ""
	outputFormat: str = ""
	sort: "bool | None" = None
	sortKeyName: "str | None" = None
	sortEncoding: "str | None" = None
	readOptions: "dict[str, Any] | None" = None
	writeOptions: "dict[str, Any] | None" = None
	sqlite: "bool | None" = None
	infoOverride: "dict[str, str] | None" = None


class GlossaryCommon(GlossaryInfo, GlossaryProgress, PluginManager):  # noqa: PLR0904

	"""
	The signature of 'convert' method is different in glossary_v2.py
		See help(Glossary.convert).

	addEntryObj is renamed to addEntry in glossary_v2.py

	These methods do not exist in glossary_v2.py (but still exist in glossary.py)

		- read():
			you can use directRead() then iterate over glossary

		- sortWords():
			you have to sort entries yourself (when adding or after directRead)

		- updateIter():
			no longer needed, and does't do anything in glossary.py

	"""

	def _closeReaders(self) -> None:
		for reader in self._readers:
			try:
				reader.close()
			except Exception:  # noqa: PERF203
				log.exception("")

	def clear(self) -> None:
		GlossaryProgress.clear(self)
		self._info = odict()

		self._data.clear()

		readers = getattr(self, "_readers", [])
		for reader in readers:
			try:
				reader.close()
			except Exception:  # noqa: PERF203
				log.exception("")
		self._readers: "list[Any]" = []
		self._defiHasWordTitle = False

		self._iter: "Iterator[EntryType] | None" = None
		self._entryFilters: "list[EntryFilterType]" = []
		self._entryFiltersExtra: "list[EntryFilterType]" = []
		self._entryFiltersName: "set[str]" = set()
		self._sort = False

		self._filename = ""
		self._defaultDefiFormat = "m"
		self._tmpDataDir = ""
		self._entryFiltersAreSet = False

	def __init__(
		self,
		info: "dict[str, str] | None" = None,
		ui: "UIType | None" = None,  # noqa: F821
	) -> None:
		"""
		info:	OrderedDict or dict instance, or None
		no need to copy OrderedDict instance before passing here
		we will not reference to it.
		"""
		GlossaryInfo.__init__(self)
		GlossaryProgress.__init__(self, ui=ui)
		self._config: "dict[str, Any]" = {}
		self._data: "EntryListType" = EntryList(
			entryToRaw=self._entryToRaw,
			entryFromRaw=self._entryFromRaw,
		)
		self._sqlite = False
		self._rawEntryCompress = False
		self._cleanupPathList: "set[str]" = set()
		self._readOptions: dict[str, Any] | None = None

		self.clear()

		if info:
			if not isinstance(info, dict | odict):
				raise TypeError(
					"Glossary: `info` has invalid type"
					", dict or OrderedDict expected",
				)
			for key, value in info.items():
				self.setInfo(key, value)

	def cleanup(self) -> None:
		self._closeReaders()
		if not self._cleanupPathList:
			return
		if not self._config.get("cleanup", True):
			log.info("Not cleaning up files:")
			log.info("\n".join(self._cleanupPathList))
			return
		self._data.close()
		for cleanupPath in self._cleanupPathList:
			if isfile(cleanupPath):
				log.debug(f"Removing file {cleanupPath}")
				try:
					os.remove(cleanupPath)
				except Exception:
					log.exception(f"error removing {cleanupPath}")
			elif isdir(cleanupPath):
				log.debug(f"Removing directory {cleanupPath}")
				rmtree(cleanupPath)
			else:
				log.error(f"no such file or directory: {cleanupPath}")
		self._cleanupPathList = set()

	def _dataEntryToRaw(self, entry: "DataEntry") -> "RawEntryType":
		b_fpath = b""
		if self.tmpDataDir:
			b_fpath = entry.save(self.tmpDataDir).encode("utf-8")
		tpl = (
			[entry.getFileName()],
			b_fpath,
			"b",
		)
		if self._rawEntryCompress:
			return zlib_compress(pickle_dumps(tpl), level=9)
		return tpl

	def _entryToRaw(self, entry: "EntryType") -> "RawEntryType":
		"""
		Return a tuple (word, defi) or (word, defi, defiFormat)
		where both word and defi might be string or list of strings.
		"""
		if entry.isData():
			return self._dataEntryToRaw(cast("DataEntry", entry))

		tpl: "tuple[list[str], bytes, str] | tuple[list[str], bytes]"
		defiFormat = entry.defiFormat
		if defiFormat and defiFormat != self._defaultDefiFormat:
			tpl = (entry.l_word, entry.b_defi, defiFormat)
		else:
			tpl = (entry.l_word, entry.b_defi)

		if self.rawEntryCompress:
			return zlib_compress(pickle_dumps(tpl), level=9)

		return tpl

	def _entryFromRaw(self, rawEntryArg: "RawEntryType") -> "EntryType":
		if isinstance(rawEntryArg, bytes):
			rawEntry = pickle_loads(zlib_decompress(rawEntryArg))
		else:
			rawEntry = rawEntryArg
		word = rawEntry[0]
		defi = rawEntry[1].decode("utf-8")
		if len(rawEntry) > 2:  # noqa: PLR2004
			defiFormat = rawEntry[2]
			if defiFormat == "b":
				fname = word
				if isinstance(fname, list):
					fname = fname[0]  # NESTED 4
				return DataEntry(fname, tmpPath=defi)
		else:
			defiFormat = self._defaultDefiFormat

		return Entry(word, defi, defiFormat=defiFormat)

	@property
	def rawEntryCompress(self) -> bool:
		return self._rawEntryCompress

	def setRawEntryCompress(self, enable: bool) -> None:
		self._rawEntryCompress = enable
		self._data.rawEntryCompress = enable

	def updateEntryFilters(self) -> None:
		entryFilters = []
		config = self._config

		glosArg = cast(GlossaryExtendedType, self)

		for configParam, default, filterClass in entryFiltersRules:
			args = []
			value = default if configParam is None else config.get(configParam, default)
			if not value:
				continue
			if not isinstance(default, bool):
				args = [value]
			entryFilters.append(filterClass(glosArg, *tuple(args)))

		if self.progressbar:
			entryFilters.append(ShowProgressBar(glosArg))

		if log.level <= core.TRACE:
			try:
				import psutil  # noqa: F401
			except ModuleNotFoundError:
				pass
			else:
				entryFilters.append(ShowMaxMemoryUsage(glosArg))

		self._entryFilters = entryFilters

		self._entryFiltersName = {entryFilter.name for entryFilter in entryFilters}

		self._entryFiltersAreSet = True

	def prepareEntryFilters(self) -> None:
		"""
		Call .prepare() method on all _entryFilters
		run this after glossary info is set and ready
		for most entry filters, it won't do anything.
		"""
		for entryFilter in self._entryFilters:
			entryFilter.prepare()

	def _addExtraEntryFilter(self, cls: "type[EntryFilterType]") -> None:
		if cls.name in self._entryFiltersName:
			return
		self._entryFilters.append(cls(cast(GlossaryType, self)))
		self._entryFiltersExtra.append(cls(cast(GlossaryType, self)))
		self._entryFiltersName.add(cls.name)

	def removeHtmlTagsAll(self) -> None:
		"""
		Remove all HTML tags from definition.

		This should only be called from a plugin's Writer.__init__ method.
		Does not apply on entries added with glos.addEntry
		"""
		self._addExtraEntryFilter(RemoveHtmlTagsAll)

	def stripFullHtml(
		self,
		errorHandler: "Callable[[EntryType, str], None] | None" = None,
	) -> None:
		"""
		Add entry filter "strip_full_html"
		to replace a full HTML document with it's body in entry definition.
		"""
		name = StripFullHtml.name
		if name in self._entryFiltersName:
			return
		self._entryFilters.append(
			StripFullHtml(
				cast(GlossaryType, self),
				errorHandler=errorHandler,
			),
		)
		self._entryFiltersName.add(name)

	def preventDuplicateWords(self) -> None:
		"""
		Add entry filter to prevent duplicate `entry.s_word`.

		This should only be called from a plugin's Writer.__init__ method.
		Does not apply on entries added with glos.addEntry

		Note: there may be still duplicate headwords or alternate words
			but we only care about making the whole `entry.s_word`
			(aka entry key) unique
		"""
		self._addExtraEntryFilter(PreventDuplicateWords)

	# def mergeEntriesWithSameHeadwordHTML(self):
	# 	"""
	# 	Merge consequtive entries that have the same word list.

	# 	Currently this convert all non-html entries to html.

	# 	Should be only called in writer.open.
	# 	"""
	# 	from pyglossary.entry_merge import mergeHtmlEntriesWithSameHeadword

	# 	self._iter = mergeHtmlEntriesWithSameHeadword(self._iter)

	def mergeEntriesWithSameHeadwordPlaintext(self):
		"""
		Merge consequtive entries that have the same word list.

		Currently this assume all entries are plaintext

		Should be only called in writer.open.
		"""
		from pyglossary.entry_merge import mergePlaintextEntriesWithSameHeadword

		self._iter = mergePlaintextEntriesWithSameHeadword(self._iter)

	def __str__(self) -> str:
		return (
			"Glossary{"
			f"filename: {self._filename!r}"
			f", name: {self._info.get('name')!r}"
			"}"
		)

	def _loadedEntryGen(self) -> "Iterator[EntryType]":
		if not self.progressbar:
			yield from self._data
			return

		filters = self._entryFiltersExtra
		if self.progressbar:
			filters.append(ShowProgressBar(cast(GlossaryExtendedType, self)))

		self.progressInit("Writing")
		for _entry in self._data:
			entry = _entry
			for f in filters:
				entry = f.run(entry)
			yield entry
		self.progressEnd()

	def _readersEntryGen(self) -> "Iterator[EntryType]":
		for reader in self._readers:
			self.progressInit("Converting")
			try:
				yield from self._applyEntryFiltersGen(reader)
			finally:
				reader.close()
			self.progressEnd()

	# This iterator/generator does not give None entries.
	# And Entry is not falsable, so bool(entry) is always True.
	# Since ProgressBar is already handled with an EntryFilter, there is
	# no point of returning None entries anymore.
	def _applyEntryFiltersGen(
		self,
		gen: "Iterator[EntryType]",
	) -> "Iterator[EntryType]":
		entry: "EntryType | None"
		for entry in gen:
			if entry is None:
				continue
			for entryFilter in self._entryFilters:
				entry = entryFilter.run(entry)  # noqa: PLW2901
				if entry is None:
					break
			else:
				yield entry

	def __iter__(self) -> "Iterator[EntryType]":
		if self._iter is not None:
			return self._iter

		if not self._readers:
			return self._loadedEntryGen()

		log.error("Glossary: iterator is not set in direct mode")
		return iter([])

	# TODO: switch to @property defaultDefiFormat
	def setDefaultDefiFormat(self, defiFormat: str) -> None:
		"""
		DefiFormat must be empty or one of these:
		"m": plain text
		"h": html
		"x": xdxf.
		"""
		self._defaultDefiFormat = defiFormat

	def getDefaultDefiFormat(self) -> str:
		return self._defaultDefiFormat

	def collectDefiFormat(
		self,
		maxCount: int,
	) -> "dict[str, float] | None":
		"""
		Collect definition format.

		Example return value:
		[("h", 0.91), ("m", 0.09)].
		"""
		from collections import Counter

		readers = self._readers
		if readers:
			log.error("collectDefiFormat: not supported in direct mode")
			return None

		counter: "dict[str, int]" = Counter()
		count = 0
		for entry in self:
			if entry.isData():
				continue
			entry.detectDefiFormat()
			counter[entry.defiFormat] += 1
			count += 1
			if count >= maxCount:
				break

		result = {
			defiFormat: itemCount / count for defiFormat, itemCount in counter.items()
		}
		for defiFormat in ("h", "m", "x"):
			if defiFormat not in result:
				result[defiFormat] = 0

		self._iter = self._loadedEntryGen()

		return result

	def __len__(self) -> int:
		return len(self._data) + sum(len(reader) for reader in self._readers)

	@property
	def config(self) -> "dict[str, Any]":
		raise NotImplementedError

	@config.setter
	def config(self, config: "dict[str, Any]") -> None:
		if self._config:
			log.error("glos.config is set more than once")
			return
		self._config = config
		if "optimize_memory" in config:
			self.setRawEntryCompress(config["optimize_memory"])

	@property
	def alts(self) -> bool:
		return self._config.get("enable_alts", True)

	@property
	def filename(self) -> str:
		return self._filename

	@property
	def tmpDataDir(self) -> str:
		if not self._tmpDataDir:
			self._setTmpDataDir(self._filename)
		return self._tmpDataDir

	@property
	def readOptions(self) -> dict | None:
		return self._readOptions

	def wordTitleStr(
		self,
		word: str,
		sample: str = "",
		_class: str = "",
	) -> str:
		"""
		Return title tag for words.

		Notes
		-----
		- `word` needs to be escaped before passing
		- `word` can contain html code (multiple words, colors, etc)
		- if input format (reader) indicates that words are already included
		in definition (as title), this method will return empty string
		- depending on glossary's `sourceLang` or writing system of `word`,
		(or sample if given) either '<b>' or '<big>' will be used.

		"""
		if self._defiHasWordTitle:
			return ""
		if not word:
			return ""
		if not sample:
			sample = word
		tag = self.titleTag(sample)
		if _class:
			return f'<{tag} class="{_class}">{word}</{tag}><br>'
		return f"<{tag}>{word}</{tag}><br>"

	def getConfig(self, name: str, default: "str | None") -> "str | None":
		return self._config.get(name, default)

	def addEntry(self, entry: "EntryType") -> None:
		self._data.append(entry)

	def newEntry(
		self,
		word: "MultiStr",
		defi: str,
		defiFormat: str = "",
		byteProgress: "tuple[int, int] | None" = None,
	) -> "Entry":
		"""
		Create and return a new entry object.

		defiFormat must be empty or one of these:
				"m": plain text
				"h": html
				"x": xdxf
		"""
		if not defiFormat:
			defiFormat = self._defaultDefiFormat

		return Entry(
			word,
			defi,
			defiFormat=defiFormat,
			byteProgress=byteProgress,
		)

	def newDataEntry(self, fname: str, data: bytes) -> "EntryType":
		import uuid

		if self._readers:
			return DataEntry(fname, data)

		if self._tmpDataDir:
			return DataEntry(
				fname,
				data,
				tmpPath=join(self._tmpDataDir, fname.replace("/", "_")),
			)

		tmpDir = join(cacheDir, "tmp")
		os.makedirs(tmpDir, mode=0o700, exist_ok=True)
		self._cleanupPathList.add(tmpDir)
		return DataEntry(
			fname,
			data,
			tmpPath=join(tmpDir, uuid.uuid1().hex),
		)

	# ________________________________________________________________________#

	# def _hasWriteAccessToDir(self, dirPath: str) -> None:
	# 	if isdir(dirPath):
	# 		return os.access(dirPath, os.W_OK)
	# 	return os.access(os.path.dirname(dirPath), os.W_OK)

	def _createReader(
		self,
		format: str,
		options: "dict[str, Any]",
	) -> "Any":
		readerClass = self.plugins[format].readerClass
		if readerClass is None:
			raise ReadError("_createReader: readerClass is None")
		reader = readerClass(self)
		for name, value in options.items():
			setattr(reader, f"_{name}", value)
		return reader

	def _setTmpDataDir(self, filename: str) -> None:
		import uuid

		# good thing about cacheDir is that we don't have to clean it up after
		# conversion is finished.
		# specially since dataEntry.save(...) will move the file from cacheDir
		# to the new directory (associated with output glossary path)
		# And we don't have to check for write access to cacheDir because it's
		# inside user's home dir. But input glossary might be in a directory
		# that we don't have write access to.
		# still maybe add a config key to decide if we should always use cacheDir
		# if self._hasWriteAccessToDir(f"{filename}_res", os.W_OK):
		# 	self._tmpDataDir = f"{filename}_res"
		# else:
		if not filename:
			filename = uuid.uuid1().hex
		self._tmpDataDir = join(cacheDir, os.path.basename(filename) + "_res")
		log.debug(f"tmpDataDir = {self._tmpDataDir}")
		os.makedirs(self._tmpDataDir, mode=0o700, exist_ok=True)
		self._cleanupPathList.add(self._tmpDataDir)

	def _validateReadoptions(
		self,
		format: str,
		options: "dict[str, Any]",
	) -> None:
		validOptionKeys = set(self.formatsReadOptions[format])
		for key in list(options):
			if key not in validOptionKeys:
				log.error(
					f"Invalid read option {key!r} given for {format} format",
				)
				del options[key]

	def _openReader(self, reader: "Any", filename: str):
		# reader.open returns "Iterator[tuple[int, int]] | None"
		progressbar: bool = self.progressbar
		try:
			openResult = reader.open(filename)
			if openResult is not None:
				self.progressInit("Reading metadata")
				lastPos = -100_000
				for pos, total in openResult:
					if progressbar and pos - lastPos > 100_000:  # noqa: PLR2004
						self.progress(pos, total, unit="bytes")
						lastPos = pos
				self.progressEnd()
		except (FileNotFoundError, LookupError) as e:
			raise ReadError(str(e)) from e

		hasTitleStr = self._info.get("definition_has_headwords", "")
		if hasTitleStr:
			if hasTitleStr.lower() == "true":
				self._defiHasWordTitle = True
			else:
				log.error(f"bad info value: definition_has_headwords={hasTitleStr!r}")

	def directRead(
		self,
		filename: str,
		format: str = "",
		**options,
	) -> bool:
		self._setTmpDataDir(filename)
		return self._read(
			filename=filename,
			format=format,
			direct=True,
			**options,
		)

	def _read(
		self,
		filename: str,
		format: str = "",
		direct: bool = False,
		**options,
	) -> bool:
		filename = os.path.abspath(filename)
		###
		inputArgs = self.detectInputFormat(filename, format=format)
		if inputArgs is None:
			return False
		origFilename = filename
		filename, format, compression = inputArgs

		if compression:
			from .compression import uncompress

			uncompress(origFilename, filename, compression)

		self._validateReadoptions(format, options)

		filenameNoExt, ext = os.path.splitext(filename)
		if ext.lower() not in self.plugins[format].extensions:
			filenameNoExt = filename

		self._filename = filenameNoExt
		if not self._info.get(c_name):
			self._info[c_name] = os.path.split(filename)[1]

		if not self._entryFiltersAreSet:
			self.updateEntryFilters()

		reader = self._createReader(format, options)
		self._openReader(reader, filename)

		self._readOptions = options

		self.prepareEntryFilters()

		if not direct:
			self.loadReader(reader)
			self._iter = self._loadedEntryGen()
			return True

		self._readers.append(reader)
		self._iter = self._readersEntryGen()

		return True

	def loadReader(self, reader: "Any") -> None:
		"""
		Iterate over `reader` object and loads the whole data into self._data
		must call `reader.open(filename)` before calling this function.
		"""
		showMemoryUsage()

		self.progressInit("Reading")
		try:
			for entry in self._applyEntryFiltersGen(reader):
				self.addEntry(entry)
		finally:
			reader.close()

		self.progressEnd()

		core.trace(log, f"Loaded {len(self._data)} entries")
		showMemoryUsage()

	def _createWriter(
		self,
		format: str,
		options: "dict[str, Any]",
	) -> "Any":
		validOptions = self.formatsWriteOptions.get(format)
		if validOptions is None:
			raise WriteError(f"No write support for {format!r} format")
		validOptionKeys = list(validOptions)
		for key in list(options):
			if key not in validOptionKeys:
				log.error(
					f"Invalid write option {key!r} given for {format} format",
				)
				del options[key]

		writerClass = self.plugins[format].writerClass
		if writerClass is None:
			raise WriteError("_createWriter: writerClass is None")
		writer = writerClass(self)
		for name, value in options.items():
			setattr(writer, f"_{name}", value)
		return writer

	def write(
		self,
		filename: str,
		format: str,
		**kwargs,
	) -> str:
		"""
		Write to a given glossary file, with given format (optional).
		Return absolute path of output file, or None if failed.

		Parameters
		----------
		filename (str): file name or path to write.
		format (str): format name
		sort (bool):
			True (enable sorting),
			False (disable sorting),
			None (auto, get from UI)
		sortKeyName (str or None):
			key function name for sorting
		sortEncoding (str or None):
			encoding for sorting, default utf-8

		You can pass write-options (of given format) as keyword arguments

		"""
		if type(filename) is not str:
			raise TypeError("filename must be str")
		if format is not None and type(format) is not str:
			raise TypeError("format must be str")

		return self._write(
			filename=filename,
			format=format,
			**kwargs,
		)

	def _writeEntries(
		self,
		writerList: "list[Any]",
		filename: str,
	) -> None:
		writer = writerList[0]
		genList = []
		gen = writer.write()
		if gen is None:
			log.error(f"{format} write function is not a generator")
		else:
			genList.append(gen)

		if self._config.get("save_info_json", False):
			from pyglossary.info_writer import InfoWriter

			infoWriter = InfoWriter(cast(GlossaryType, self))
			filenameNoExt, _, _, _ = splitFilenameExt(filename)
			infoWriter.open(f"{filenameNoExt}.info")
			genList.append(infoWriter.write())
			writerList.append(infoWriter)

		for gen in genList:
			gen.send(None)
		for entry in self:
			for gen in genList:
				gen.send(entry)
		# suppress() on the whole for-loop does not work
		for gen in genList:
			with suppress(StopIteration):
				gen.send(None)

	@staticmethod
	def _openWriter(
		writer: "Any",
		filename: str,
	):
		try:
			writer.open(filename)
		except (FileNotFoundError, LookupError) as e:
			raise WriteError(str(e)) from e

	def _write(
		self,
		filename: str,
		format: str,
		sort: bool = False,
		**options,
	) -> str:
		filename = os.path.abspath(filename)

		if format not in self.plugins or not self.plugins[format].canWrite:
			raise WriteError(f"No Writer class found for plugin {format}")

		if self._readers and sort:
			log.warning(
				"Full sort enabled, falling back to indirect mode",
			)
			for reader in self._readers:
				self.loadReader(reader)
			self._readers = []

		log.info(f"Writing to {format} file {filename!r}")

		writer = self._createWriter(format, options)

		self._sort = sort

		if sort:
			t0 = now()
			self._data.sort()
			log.info(f"Sorting took {now() - t0:.1f} seconds")

		if self._readers:
			self._iter = self._readersEntryGen()
		else:
			self._iter = self._loadedEntryGen()
		self._openWriter(writer, filename)

		showMemoryUsage()

		writerList = [writer]
		try:
			self._writeEntries(writerList, filename)
		except (FileNotFoundError, LookupError) as e:
			raise WriteError(str(e)) from e
		finally:
			showMemoryUsage()
			log.debug("Running writer.finish()")
			for writer in writerList:
				writer.finish()
			self.clear()

		showMemoryUsage()

		return filename

	def _compressOutput(self, filename: str, compression: str) -> str:
		from .compression import compress

		return compress(
			cast(GlossaryType, self),
			filename,
			compression,
		)

	def _switchToSQLite(
		self,
		inputFilename: str,
	) -> None:
		from .sq_entry_list import SqEntryList

		sq_fpath = join(cacheDir, f"{os.path.basename(inputFilename)}.db")
		if isfile(sq_fpath):
			log.info(f"Removing and re-creating {sq_fpath!r}")
			os.remove(sq_fpath)

		self._data = SqEntryList(
			entryToRaw=self._entryToRaw,
			entryFromRaw=self._entryFromRaw,
			filename=sq_fpath,
			create=True,
			persist=True,
		)
		self._cleanupPathList.add(sq_fpath)

		if not self.alts:
			log.warning(
				"SQLite mode only works with enable_alts=True, force-enabling it.",
			)
		self._config["enable_alts"] = True
		self._sqlite = True

	@staticmethod
	def _checkSortFlag(
		plugin: "PluginProp",
		sort: "bool | None",
	) -> bool:
		sortOnWrite = plugin.sortOnWrite
		if sortOnWrite == ALWAYS:
			if sort is False:
				log.warning(
					f"Writing {plugin.name} requires sorting"
					", ignoring user sort=False option",
				)
			return True

		if sortOnWrite == NEVER:
			if sort:
				log.warning(
					"Plugin prevents sorting before write"
					", ignoring user sort=True option",
				)
			return False

		if sortOnWrite == DEFAULT_YES:
			return sort or sort is None

		# if sortOnWrite == DEFAULT_NO:
		return bool(sort)

	def _resolveSortParams(
		self,
		args: ConvertArgs,
		plugin: "PluginProp",
	) -> "tuple[bool, bool]":
		"""
		sortKeyName: see doc/sort-key.md.

		returns (sort, direct)
		"""
		if args.direct and args.sqlite:
			raise ValueError(
				f"Conflictng arguments: direct={args.direct}, sqlite={args.sqlite}",
			)

		sort = self._checkSortFlag(plugin, args.sort)

		if not sort:
			if args.direct is None:
				return True, False
			return args.direct, False

		# from this point we can assume sort == True and direct == False

		sqlite = args.sqlite
		if sqlite is None:
			sqlite = self._config.get("auto_sqlite", True)
			if sqlite:
				log.info(
					"Automatically switching to SQLite mode"
					f" for writing {plugin.name}",
				)

		sortKeyTuple = self._checkSortKey(
			plugin,
			args.sortKeyName,
			args.sortEncoding,
		)
		namedSortKey, sortEncoding = sortKeyTuple

		if sqlite:
			self._switchToSQLite(
				inputFilename=args.inputFilename,
			)

		self._data.setSortKey(
			namedSortKey=namedSortKey,
			sortEncoding=sortEncoding,
			writeOptions=args.writeOptions or {},
		)

		return False, True

	@staticmethod
	def _checkSortKey(
		plugin: "PluginProp",
		sortKeyName: "str | None",
		sortEncoding: "str | None",
	) -> "tuple[NamedSortKey, str]":
		"""
		Check sortKeyName, sortEncoding and (output) plugin's params
		returns (namedSortKey, sortEncoding).
		"""
		writerSortKeyName = plugin.sortKeyName
		writerSortEncoding = getattr(plugin, "sortEncoding", None)
		if plugin.sortOnWrite == ALWAYS:
			if not writerSortKeyName:
				raise Error("No sortKeyName was found in plugin")

			if sortKeyName and sortKeyName != writerSortKeyName:
				log.warning(
					f"Ignoring user-defined sort order {sortKeyName!r}"
					f", and using sortKey function from {plugin.name} plugin",
				)
			sortKeyName = writerSortKeyName

			if writerSortEncoding:
				sortEncoding = writerSortEncoding
		elif not sortKeyName:
			sortKeyName = writerSortKeyName or defaultSortKeyName

		namedSortKey = lookupSortKey(sortKeyName)
		if namedSortKey is None:
			raise Error(f"invalid {sortKeyName = }")

		log.info(f"Using sortKeyName = {namedSortKey.name!r}")

		if not sortEncoding:
			sortEncoding = "utf-8"

		return namedSortKey, sortEncoding

	@staticmethod
	def _convertValidateStrings(args: ConvertArgs) -> None:
		if type(args.inputFilename) is not str:
			raise TypeError("inputFilename must be str")
		if type(args.outputFilename) is not str:
			raise TypeError("outputFilename must be str")

		if args.inputFormat is not None and type(args.inputFormat) is not str:
			raise TypeError("inputFormat must be str")
		if args.outputFormat is not None and type(args.outputFormat) is not str:
			raise TypeError("outputFormat must be str")

	def _convertPrepare(
		self,
		args: ConvertArgs,
		outputFilename: str = "",
		outputFormat: str = "",
	) -> "bool | None":
		if isdir(outputFilename) and os.listdir(outputFilename):
			raise Error(
				f"Directory already exists and not empty: {relpath(outputFilename)}",
			)

		outputPlugin = self.plugins[outputFormat]

		sortParams = self._resolveSortParams(
			args=args,
			plugin=outputPlugin,
		)
		direct, sort = sortParams

		showMemoryUsage()

		self._setTmpDataDir(args.inputFilename)

		readOptions = args.readOptions or {}

		self._read(
			args.inputFilename,
			format=args.inputFormat,
			direct=direct,
			**readOptions,
		)

		self.detectLangsFromName()

		return sort

	def convertV2(self, args: ConvertArgs) -> str:
		"""
		Return absolute path of output file, or None if failed.

		sortKeyName:
			name of sort key/algorithm
			defaults to `defaultSortKeyName` in sort_keys.py
			see doc/sort-key.md or sort_keys.py for other possible values
			This can also include sort locale after a colon sign, for example:
				sortKeyName=":fa_IR.UTF-8"
				sortKeyName="headword:fa_IR.UTF-8"

		sortEncoding:
			encoding/charset for sorting, default to utf-8
		"""
		self._convertValidateStrings(args)
		if args.outputFilename == args.inputFilename:
			raise Error("Input and output files are the same")

		tm0 = now()

		outputArgs = self.detectOutputFormat(
			filename=args.outputFilename,
			format=args.outputFormat,
			inputFilename=args.inputFilename,
		)
		if not outputArgs:
			raise Error(f"Writing file {relpath(args.outputFilename)!r} failed.")
		outputFilename, outputFormat, compression = outputArgs

		sort = self._convertPrepare(
			args=args,
			outputFilename=outputFilename,
			outputFormat=outputFormat,
		)
		if sort is None:
			return None

		if args.infoOverride:
			for key, value in args.infoOverride.items():
				self.setInfo(key, value)

		if compression and not self.plugins[outputFormat].singleFile:
			os.makedirs(outputFilename, mode=0o700, exist_ok=True)

		writeOptions = args.writeOptions or {}

		finalOutputFile = self._write(
			outputFilename,
			outputFormat,
			sort=sort,
			**writeOptions,
		)

		if compression:
			finalOutputFile = self._compressOutput(finalOutputFile, compression)

		log.info(f"Writing file {finalOutputFile!r} done.")
		log.info(f"Running time of convert: {now() - tm0:.1f} seconds")
		showMemoryUsage()
		self.cleanup()

		return finalOutputFile


# ________________________________________________________________________#


class Glossary(GlossaryCommon):

	"""
	init method is inherited from PluginManager
		arguments:
			usePluginsJson: bool = True
			skipDisabledPlugins: bool = True.

		init() must be called only once, so make sure you put it in the
		right place. Probably in the top of your program's main function or module.
	"""

	GLOSSARY_API_VERSION = "2.0"

	def convert(self, args: ConvertArgs) -> "str | None":
		return self.convertV2(args)



# -*- coding: utf-8 -*-
# mypy: ignore-errors
#
# Copyright © 2008-2021 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Copyright © 2011-2012 kubtek <kubtek@gmail.com>
# This file is part of PyGlossary project, http://github.com/ilius/pyglossary
# Thanks to Raul Fernandes <rgfbr@yahoo.com.br> and Karl Grill for reverse
# engineering as part of https://sourceforge.net/projects/ktranslator/
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

from collections.abc import Callable
from typing import Any

from pyglossary import gregorian
from pyglossary.core import log
from pyglossary.text_utils import (
	uintFromBytes,
)

from .bgl_charset import charsetByCode
from .bgl_language import BabylonLanguage, languageByCode

__all__ = ["charsetInfoDecode", "infoType3ByCode"]


class InfoItem:
	__slots__ = (
		"attr",
		"decode",
		"name",
	)

	def __init__(
		self,
		name: str,
		decode: "Callable[[bytes], Any] | None" = None,
		attr: bool = False,
	) -> None:
		self.name = name
		self.decode = decode
		self.attr = attr


def decodeBglBinTime(b_value: bytes) -> str:
	jd1970 = gregorian.to_jd(1970, 1, 1)
	djd, hm = divmod(uintFromBytes(b_value), 24 * 60)
	year, month, day = gregorian.jd_to(djd + jd1970)
	hour, minute = divmod(hm, 60)
	return f"{year:04d}/{month:02d}/{day:02d}, {hour:02d}:{minute:02d}"


def languageInfoDecode(b_value: bytes) -> "BabylonLanguage | None":
	"""Returns BabylonLanguage instance."""
	intValue = uintFromBytes(b_value)
	try:
		return languageByCode[intValue]
	except IndexError:
		log.warning(f"read_type_3: unknown language code = {intValue}")
		return None


def charsetInfoDecode(b_value: bytes) -> "str | None":
	value = b_value[0]
	try:
		return charsetByCode[value]
	except KeyError:
		log.warning(f"read_type_3: unknown charset {value!r}")
	return None


def aboutInfoDecode(b_value: bytes) -> "dict[str, any]":
	if not b_value:
		return None
	b_aboutExt, _, aboutContents = b_value.partition(b"\x00")
	if not b_aboutExt:
		log.warning("read_type_3: about: no file extension")
		return None
	try:
		aboutExt = b_aboutExt.decode("ascii")
	except UnicodeDecodeError as e:
		log.error(f"{b_aboutExt=}: {e}")
		aboutExt = ""
	return {
		"about_extension": aboutExt,
		"about": aboutContents,
	}


def utf16InfoDecode(b_value: bytes) -> "str | None":
	r"""
	Decode info values from UTF-16.

	Return str, or None (on errors).

	block type = 3
	block format: <2 byte code1><2 byte code2>
	if code2 == 0: then the block ends
	if code2 == 1: then the block continues as follows:
	<4 byte len1> \x00 \x00 <message in utf-16>
	len1 - length of message in 2-byte chars
	"""
	if b_value[0] != 0:
		log.warning(
			f"utf16InfoDecode: b_value={b_value}, null expected at 0",
		)
		return None

	if b_value[1] == 0:
		if len(b_value) > 2:
			log.warning(
				f"utf16InfoDecode: unexpected b_value size: {len(b_value)}",
			)
		return None

	if b_value[1] > 1:
		log.warning(
			f"utf16InfoDecode: b_value={b_value!r}, unexpected byte at 1",
		)
		return None

	# now b_value[1] == 1
	size = 2 * uintFromBytes(b_value[2:6])
	if tuple(b_value[6:8]) != (0, 0):
		log.warning(
			f"utf16InfoDecode: b_value={b_value!r}, null expected at 6:8",
		)
	if size != len(b_value) - 8:
		log.warning(
			f"utf16InfoDecode: b_value={b_value!r}, size does not match",
		)

	return b_value[8:].decode("utf16")  # str


def flagsInfoDecode(b_value: bytes) -> "dict[str, bool]":
	"""
	Returns a dict with these keys:
	utf8Encoding
	when this flag is set utf8 encoding is used for all articles
	when false, the encoding is set according to the source and
	target alphabet
	bgl_spellingAlternatives
	determines whether the glossary offers spelling alternatives
	for searched terms
	bgl_caseSensitive
	defines if the search for terms in this glossary is
	case sensitive
	see code 0x20 as well.

	"""
	flags = uintFromBytes(b_value)
	return {
		"utf8Encoding": (flags & 0x8000 != 0),
		"bgl_spellingAlternatives": (flags & 0x10000 == 0),
		"bgl_caseSensitive": (flags & 0x1000 != 0),
	}


infoType3ByCode = {
	# glossary name
	0x01: InfoItem("title"),
	# glossary author name, a list of "|"-separated values
	0x02: InfoItem("author"),
	# glossary author e-mail
	0x03: InfoItem("email"),
	0x04: InfoItem("copyright"),
	0x07: InfoItem(
		"sourceLang",
		decode=languageInfoDecode,
		attr=True,
	),
	0x08: InfoItem(
		"targetLang",
		decode=languageInfoDecode,
		attr=True,
	),
	0x09: InfoItem("description"),
	# 0: browsing disabled, 1: browsing enabled
	0x0A: InfoItem(
		"bgl_browsingEnabled",
		decode=lambda b_value: (b_value[0] != 0),
	),
	0x0B: InfoItem("icon1.ico"),
	0x0C: InfoItem(
		"bgl_numEntries",
		decode=uintFromBytes,
		attr=True,
	),
	# the value is a dict
	0x11: InfoItem("flags", decode=flagsInfoDecode),
	0x14: InfoItem("creationTime", decode=decodeBglBinTime),
	0x1A: InfoItem(
		"sourceCharset",
		decode=charsetInfoDecode,
		attr=True,
	),
	0x1B: InfoItem(
		"targetCharset",
		decode=charsetInfoDecode,
		attr=True,
	),
	0x1C: InfoItem(
		"bgl_firstUpdated",
		decode=decodeBglBinTime,
	),
	# bgl_firstUpdated was previously called middleUpdated
	# in rare cases, bgl_firstUpdated is before creationTime
	# but usually it looks like to be the first update (after creation)
	# in some cases, it's the same as lastUpdated
	# in some cases, it's minutes after creationTime
	# bgl_firstUpdated exists in more glossaries than lastUpdated
	# so if lastUpdated is not there, we use bgl_firstUpdated as lastUpdated
	0x20: InfoItem(
		"bgl_caseSensitive2",
		decode=lambda b_value: (b_value[0] == 0x31),
		# 0x30 - case sensitive search is disabled
		# 0x31 - case sensitive search is enabled
	),
	0x24: InfoItem("icon2.ico"),
	0x2C: InfoItem(
		"bgl_purchaseLicenseMsg",
		decode=utf16InfoDecode,
	),
	0x2D: InfoItem(
		"bgl_licenseExpiredMsg",
		decode=utf16InfoDecode,
	),
	0x2E: InfoItem("bgl_purchaseAddress"),
	0x30: InfoItem(
		"bgl_titleWide",
		decode=utf16InfoDecode,
	),
	# a list of "|"-separated values
	0x31: InfoItem(
		"bgl_authorWide",
		decode=utf16InfoDecode,
	),
	0x33: InfoItem(
		"lastUpdated",
		decode=decodeBglBinTime,
	),
	0x3B: InfoItem("bgl_contractions"),
	# contains a value like "Arial Unicode MS" or "Tahoma"
	0x3D: InfoItem("bgl_fontName"),
	# value would be dict
	0x41: InfoItem(
		"bgl_about",
		decode=aboutInfoDecode,
	),
	# the length of the substring match in a term
	0x43: InfoItem(
		"bgl_length",
		decode=uintFromBytes,
	),
}



# -*- coding: utf-8 -*-

import html
import os
import re
import time
from collections.abc import Generator
from functools import lru_cache
from os.path import isdir, isfile, join
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	import io

from pyglossary.core import log
from pyglossary.glossary_types import (
	EntryType,
	GlossaryType,
)
from pyglossary.option import (
	BoolOption,
	EncodingOption,
	IntOption,
	Option,
	StrOption,
)
from pyglossary.text_utils import (
	escapeNTB,
	unescapeNTB,
)

__all__ = [
	"Writer",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

enable = True
lname = "html_dir"
format = "HtmlDir"
description = "HTML Directory"
extensions = (".hdir",)
extensionCreate = ".hdir/"
singleFile = False
kind = "directory"
wiki = ""
website = None
optionsProp: "dict[str, Option]" = {
	"encoding": EncodingOption(),
	"resources": BoolOption(
		comment="Enable resources / data files",
	),
	"max_file_size": IntOption(
		comment="Maximum file size in bytes",
	),
	"filename_format": StrOption(
		comment="Filename format, default: {n:05d}.html",
	),
	"escape_defi": BoolOption(
		comment="Escape definitions",
	),
	"dark": BoolOption(
		comment="Use dark style",
	),
	"css": StrOption(
		comment="Path to css file",
	),
	"word_title": BoolOption(
		comment="Add headwords title to beginning of definition",
	),
}

nbsp = "\xa0"
# nbsp = "&nbsp;"

darkStyle = """
body {{
	background-color: #373737;
	color: #eee;
}}
a {{ color: #aaaaff; }}
a.broken {{ color: #e0c0c0; }}
a.no_ul {{ text-decoration: none; }}
b.headword {{ font-size: 1.5em; color: #c7ffb9; }}
h1 {{ font-size: 1.5em; color: #c7ffb9;}}
h2 {{ font-size: 1.3em;}}
h3 {{ font-size: 1.0em;}}
h4 {{ font-size: 1.0em;}}
h5 {{ font-size: 1.0em;}}
h6 {{ font-size: 1.0em;}}
"""


class Writer:
	_encoding: str = "utf-8"
	_resources: bool = True
	_max_file_size: int = 102400
	_filename_format: str = "{n:05d}.html"
	_escape_defi: bool = False
	_dark: bool = True
	_css: str = ""
	_word_title: bool = True

	@staticmethod
	def stripFullHtmlError(entry: "EntryType", error: str) -> None:
		log.error(f"error in stripFullHtml: {error}, words={entry.l_word!r}")

	def __init__(self, glos: GlossaryType) -> None:
		self._glos = glos
		self._filename = ""
		self._fileObj: "io.IOBase | None" = None
		self._encoding = "utf-8"
		self._filename_format = "{n:05d}.html"
		self._tail = "</body></html>"
		self._filenameList: "list[str]" = []
		glos.stripFullHtml(errorHandler=self.stripFullHtmlError)

		self._resSrcPattern = re.compile(' src="([^"]*)"')

	def open(self, filename: str) -> None:
		self._filename = filename
		self._resDir = resDir = join(filename, "res")
		if not isdir(filename):
			os.mkdir(filename)
		if not isdir(resDir):
			os.mkdir(resDir)
		if self._css:
			self.copyCSS(self._css)

	def copyCSS(self, cssPath: str) -> None:
		import shutil

		shutil.copy(cssPath, join(self._filename, "style.css"))

	def finish(self) -> None:
		pass

	def getNextFilename(self) -> str:
		return self._filename_format.format(
			n=len(self._filenameList),
		)

	def nextFile(self) -> "io.TextIOBase":
		if self._fileObj:
			self._fileObj.write(self._tail)
			self._fileObj.close()
		filename = self.getNextFilename()
		self._filenameList.append(filename)
		self._fileObj = open(
			join(
				self._filename,
				filename,
			),
			mode="w",
			encoding=self._encoding,
		)
		return self._fileObj

	def fixLinks(self, linkTargetSet: "set[str]") -> None:  # noqa: PLR0912
		import gc

		gc.collect()
		dirn = self._filename

		filenameList = self._filenameList

		fileByWord: "dict[str, list[tuple[str, int]]]" = {}
		for line in open(join(dirn, "index.txt"), encoding="utf-8"):
			line = line.rstrip("\n")  # noqa: PLW2901
			if not line:
				continue
			entryIndexStr, wordEsc, filename, _ = line.split("\t")
			entryIndex = int(entryIndexStr)
			# entryId = f"entry{entryIndex}"
			word = unescapeNTB(wordEsc)
			if word not in linkTargetSet:
				continue
			if word in fileByWord:
				fileByWord[word].append((filename, entryIndex))
			else:
				fileByWord[word] = [(filename, entryIndex)]

		# with open(join(dirn, "fileByWord.json"), "w") as fileByWordFile:
		# 	json.dump(fileByWord, fileByWordFile, ensure_ascii=False, indent="\t")

		@lru_cache(maxsize=10)
		def getLinksByFile(fileIndex: int) -> "io.TextIOBase":
			return open(
				join(dirn, f"links{fileIndex}"),
				mode="a",
				encoding="utf-8",
			)

		log.info("")
		for line in open(join(dirn, "links.txt"), encoding="utf-8"):
			line = line.rstrip("\n")  # noqa: PLW2901
			if not line:
				continue
			target, fileIndexStr, x_start, x_size = line.split("\t")
			target = unescapeNTB(target)
			if target not in fileByWord:
				targetNew = ""
			else:
				targetFilename, targetEntryIndex = fileByWord[target][0]
				if targetFilename == filename:
					continue
				targetNew = f"{targetFilename}#entry{targetEntryIndex}"
			_file = getLinksByFile(int(fileIndexStr))
			_file.write(
				f"{x_start}\t{x_size}\t{targetNew}\n",
			)
			_file.flush()

		linkTargetSet.clear()
		del fileByWord, linkTargetSet
		gc.collect()

		if os.sep == "\\":
			time.sleep(0.1)

		entry_url_fmt = self._glos.getInfo("entry_url")

		re_href = re.compile(
			b' href="[^<>"]*?"',
			re.IGNORECASE,
		)

		for fileIndex, filename in enumerate(filenameList):
			if not isfile(join(dirn, f"links{fileIndex}")):
				continue
			with open(join(dirn, filename), mode="rb") as inFile:
				with open(join(dirn, f"{filename}.new"), mode="wb") as outFile:
					for linkLine in open(join(dirn, f"links{fileIndex}"), "rb"):
						outFile.flush()
						(
							b_x_start,
							b_x_size,
							b_target,
						) = linkLine.rstrip(b"\n").split(b"\t")
						outFile.write(
							inFile.read(
								int(b_x_start, 16) - inFile.tell(),
							),
						)
						curLink = inFile.read(int(b_x_size, 16))

						if b_target:
							outFile.write(
								re_href.sub(
									b' href="./' + b_target + b'"',
									curLink,
								),
							)
							continue

						if not entry_url_fmt:
							outFile.write(
								curLink.replace(
									b' href="#',
									b' class="broken" href="#',
								),
							)
							continue

						_st = curLink.decode("utf-8")
						i = _st.find('href="#')
						j = _st.find('"', i + 7)
						word = _st[i + 7 : j]
						url = entry_url_fmt.format(word=word)
						outFile.write(
							(
								_st[:i] + f'class="broken" href="{url}"' + _st[j + 1 :]
							).encode("utf-8"),
						)

					outFile.write(inFile.read())

			os.remove(join(dirn, filename))
			os.rename(join(dirn, f"{filename}.new"), join(dirn, filename))
			os.remove(join(dirn, f"links{fileIndex}"))

	def writeInfo(self, filename: str, header: str) -> None:
		glos = self._glos
		title = glos.getInfo("name")
		customStyle = (
			"table, th, td {border: 1px solid black; "
			"border-collapse: collapse; padding: 5px;}"
		)
		infoHeader = header.format(
			pageTitle=f"Info: {title}",
			customStyle=customStyle,
		)
		with open(
			join(filename, "info.html"),
			mode="w",
			encoding=self._encoding,
		) as _file:
			_file.write(
				infoHeader + "<table>"
				"<tr>"
				'<th width="%10">Key</th>'
				'<th width="%90">Value</th>'
				"</tr>\n",
			)
			for key, value in glos.iterInfo():
				_file.write(
					f"<tr><td>{key}</td><td>{value}</td></tr>\n",
				)
			_file.write("</table></body></html>")

	@staticmethod
	def _subResSrc(m: "re.Match") -> str:
		url = m.group(1)
		if "://" in url:
			return m.group(0)
		url = "res/" + url
		return f' src="{url}"'

	def write(self) -> "Generator[None, EntryType, None]":  # noqa: PLR0912
		encoding = self._encoding
		resources = self._resources
		max_file_size = self._max_file_size
		filename_format = self._filename_format
		escape_defi = self._escape_defi

		wordSep = ' <font color="red">|</font> '

		initFileSizeMax = 100

		glos = self._glos

		filename = self._filename
		self._encoding = encoding
		self._filename_format = filename_format

		entry_url_fmt = glos.getInfo("entry_url")

		def getEntryWebLink(entry: "EntryType") -> str:
			if not entry_url_fmt:
				return ""
			url = entry_url_fmt.format(word=html.escape(entry.l_word[0]))
			return f'{nbsp}<a class="no_ul" href="{url}">&#127759;</a>'

		# from math import log2, ceil
		# maxPosHexLen = int(ceil(log2(max_file_size) / 4))

		indexTxtFileObj = open(
			join(filename, "index.txt"),
			mode="w",
			encoding="utf-8",
		)
		linksTxtFileObj = open(
			join(filename, "links.txt"),
			mode="w",
			encoding="utf-8",
		)

		title = glos.getInfo("name")
		style = ""
		if self._dark:
			style = darkStyle

		cssLink = '<link rel="stylesheet" href="style.css" />' if self._css else ""

		header = (
			"<!DOCTYPE html>\n"
			"<html><head>"
			"<title>{pageTitle}</title>"
			f'<meta charset="{encoding}">'
			f'<style type="text/css">{style}{{customStyle}}</style>{cssLink}'
			"</meta></head><body>\n"
		)

		def pageHeader(n: int) -> str:
			return header.format(
				pageTitle=f"Page {n} of {title}",
				customStyle="",
			)

		def navBar() -> str:
			links = []
			if len(self._filenameList) > 1:
				links.append(f'<a href="./{self._filenameList[-2]}">&#9664;</a>')
			links.extend(
				[
					f'<a href="./{self.getNextFilename()}">&#9654;</a>',
					'<a href="./info.html">ℹ️</a></div>',  # noqa: RUF001
				],
			)
			return (
				'<nav style="text-align: center; font-size: 2.5em;">'
				+ f"{nbsp}{nbsp}{nbsp}".join(links)
				+ "</nav>"
			)

		tailSize = len(self._tail.encode(encoding))

		if max_file_size < len(header) + tailSize:
			raise ValueError(f"{max_file_size=} is too small")

		max_file_size -= tailSize

		if not isdir(self._filename):
			os.mkdir(self._filename)

		fileObj = self.nextFile()
		fileObj.write(pageHeader(0))
		fileObj.write(navBar())

		re_fixed_link = re.compile(
			r'<a (?:[^<>]*? )?href="#([^<>"]+?)">[^<>]+?</a>',
			re.IGNORECASE,
		)

		linkTargetSet = set()

		def replaceBword(text: str) -> str:
			return text.replace(
				' href="bword://',
				' href="#',
			)

		def addLinks(text: str, pos: int) -> None:
			for m in re_fixed_link.finditer(text):
				if ' class="entry_link"' in m.group(0):
					continue
				if m.group(0).count("href=") != 1:
					log.error(f"unexpected match: {m.group(0)}")
				target = html.unescape(m.group(1))
				linkTargetSet.add(target)
				start = m.start()
				b_start = len(text[:start].encode(encoding))
				b_size = len(text[start : m.end()].encode(encoding))
				linksTxtFileObj.write(
					f"{escapeNTB(target)}\t"
					f"{len(self._filenameList) - 1}\t"
					f"{pos + b_start:x}\t"
					f"{b_size:x}\n",
				)
				linksTxtFileObj.flush()

		self.writeInfo(filename, header)

		_word_title = self._word_title

		resDir = self._resDir
		entryIndex = -1
		while True:
			entryIndex += 1
			entry = yield
			if entry is None:
				break
			if entry.isData():
				if resources:
					entry.save(resDir)
				continue

			entry.detectDefiFormat()
			defi = entry.defi
			defiFormat = entry.defiFormat

			if defi.startswith("<!DOCTYPE html>") and defiFormat != "h":
				log.error(f"bad {defiFormat=}")
				defiFormat = "h"

			if defiFormat == "m":
				defi = html.escape(defi)
				if "\n" in defi:
					# could be markdown or unformatted plaintext
					# FIXME: this changes the font to a monospace
					defi = f"<pre>{defi}</pre>"
			elif defiFormat == "h":
				defi = self._resSrcPattern.sub(self._subResSrc, defi)
				if escape_defi:
					defi = html.escape(defi)

			entryId = f"entry{entryIndex}"

			if _word_title:
				words = [html.escape(word) for word in entry.l_word]
				title = glos.wordTitleStr(
					wordSep.join(words),
					sample=entry.l_word[0],
					_class="headword",
				)

			if not title:
				title = f"Entry {entryIndex}"

			# entry_link_sym = "&#182;"
			entry_link_sym = "&#128279;"
			text = (
				f'<div id="{entryId}">{title}{nbsp}{nbsp}'
				f'<a class="no_ul" class="entry_link" href="#{entryId}">'
				f"{entry_link_sym}</a>"
				f"{getEntryWebLink(entry)}"
				f"<br>\n{defi}"
				"</div>\n"
				"<hr>\n"
			)
			pos = fileObj.tell()
			if pos > initFileSizeMax and pos > max_file_size - len(
				text.encode(encoding),
			):
				fileObj = self.nextFile()
				fileObj.write(
					pageHeader(
						len(self._filenameList) - 1,
					),
				)
				fileObj.write(navBar())
			pos = fileObj.tell()
			tmpFilename = escapeNTB(self._filenameList[-1])
			for word in entry.l_word:
				indexTxtFileObj.write(
					f"{entryIndex}\t"
					f"{escapeNTB(word)}\t"
					f"{tmpFilename}\t"
					f"{pos}\n",
				)
			del tmpFilename
			text = replaceBword(text)
			addLinks(text, pos)
			fileObj.write(text)

		fileObj.close()
		self._fileObj = None
		indexTxtFileObj.close()

		linksTxtFileObj.close()

		if linkTargetSet:
			log.info(f"{len(linkTargetSet)} link targets found")
			log.info("Fixing links, please wait...")
			self.fixLinks(linkTargetSet)

		os.remove(join(filename, "links.txt"))



from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from .sort_keys_types import sortKeyType, sqliteSortKeyType


desc = "ASCII-Lowercase Headword"


def normal(
	sortEncoding: str = "utf-8",
	**_options,
) -> "sortKeyType":
	def sortKey(words: "list[str]") -> bytes:
		return words[0].encode(sortEncoding, errors="replace").lower()

	return sortKey


# def locale(
# 	collator: "T_Collator",  # noqa: F821
# ) -> "sortKeyType":
# 	raise NotImplementedError("")


def sqlite(sortEncoding: str = "utf-8", **_options) -> "sqliteSortKeyType":
	def sortKey(words: "list[str]") -> bytes:
		return words[0].encode(sortEncoding, errors="replace").lower()

	return [
		(
			"headword_blower",
			"TEXT" if sortEncoding == "utf-8" else "BLOB",
			sortKey,
		),
	]



from collections.abc import Callable
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from pyglossary.icu_types import T_Collator

	from .sort_keys_types import sortKeyType, sqliteSortKeyType


desc = "Lowercase Headword"


def normal(sortEncoding: str = "utf-8", **_options) -> "sortKeyType":
	def sortKey(words: "list[str]") -> bytes:
		return words[0].lower().encode(sortEncoding, errors="replace")

	return sortKey


def locale(
	collator: "T_Collator",  # noqa: F821
) -> "sortKeyType":
	cSortKey = collator.getSortKey

	def sortKey(words: "list[str]") -> bytes:
		return cSortKey(words[0].lower())

	return lambda **_options: sortKey


def sqlite(
	sortEncoding: str = "utf-8",
	**_options,
) -> "sqliteSortKeyType":
	def sortKey(words: "list[str]") -> bytes:
		return words[0].lower().encode(sortEncoding, errors="replace")

	return [
		(
			"headword_lower",
			"TEXT" if sortEncoding == "utf-8" else "BLOB",
			sortKey,
		),
	]


def sqlite_locale(
	collator: "T_Collator",  # noqa: F821
) -> "Callable[..., sqliteSortKeyType]":
	cSortKey = collator.getSortKey

	def sortKey(words: "list[str]") -> bytes:
		return cSortKey(words[0].lower())

	return lambda **_options: [("sortkey", "BLOB", sortKey)]



# -*- coding: utf-8 -*-
# appledict/_normalize.py
# Output to Apple Dictionary xml sources for Dictionary Development Kit.
#
# Copyright © 2016-2019 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Copyright © 2016 ivan tkachenko me@ratijas.tk
# Copyright © 2012-2015 Xiaoqiang Wang <xiaoqiangwang AT gmail DOT com>
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

import re
from typing import Any

re_spaces = re.compile(r"[ \t\n]{2,}")
re_title = re.compile('<[^<]+?>|"|[<>]|\xef\xbb\xbf')
re_title_short = re.compile(r"\[.*?\]")
re_whitespace = re.compile("(\t|\n|\r)")

# FIXME: rename all/most functions here, add a 'fix_' prefix


def spaces(s: str) -> str:
	"""
	Strip off leading and trailing whitespaces and
	replace contiguous whitespaces with just one space.
	"""
	return re_spaces.sub(" ", s.strip())


_brackets_sub = (
	(
		re.compile(r"( *)\{( *)\\\[( *)"),  # { \[
		r"\1\2\3[",
	),
	(
		re.compile(r"( *)\\\]( *)\}( *)"),  # \] }
		r"]\1\2\3",
	),
	(
		re.compile(r"( *)\{( *)\(( *)\}( *)"),  # { ( }
		r"\1\2\3\4[",
	),
	(
		re.compile(r"( *)\{( *)\)( *)\}( *)"),  # { ) }
		r"]\1\2\3\4",
	),
	(
		re.compile(r"( *)\{( *)\(( *)"),  # { (
		r"\1\2\3[",
	),
	(
		re.compile(r"( *)\)( *)\}( *)"),  # ) }
		r"]\1\2\3",
	),
	(
		re.compile(r"( *)\{( *)"),  # {
		r"\1\2[",
	),
	(
		re.compile(r"( *)\}( *)"),  # }
		r"]\1\2",
	),
	(
		re.compile(r"{.*?}"),
		r"",
	),
)


def brackets(s: str) -> str:
	r"""
	Replace all crazy brackets with square ones [].

	following combinations are to replace:
		{ \[ ... \] }
		{ ( } ... { ) }
		{ ( ... ) }
		{ ... }
	"""
	if "{" in s:
		for exp, sub in _brackets_sub:
			s = exp.sub(sub, s)
	return spaces(s)


def truncate(text: str, length: int = 449) -> str:
	"""
	Trunct a string to given length
	:param str text:
	:return: truncated text
	:rtype: str.
	"""
	content = re_whitespace.sub(" ", text)
	if len(text) > length:
		# find the next space after max_len chars (do not break inside a word)
		pos = content[:length].rfind(" ")
		if pos == -1:
			pos = length
		text = text[:pos]
	return text  # noqa: RET504


def title(title: str, BeautifulSoup: "Any") -> str:
	"""Strip double quotes and html tags."""
	if BeautifulSoup:
		title = title.replace("\xef\xbb\xbf", "")
		if len(title) > 1:
			# BeautifulSoup has a bug when markup <= 1 char length
			title = BeautifulSoup.BeautifulSoup(
				title,
				features="lxml",
				# FIXME: html or lxml? gives warning unless it's lxml
			).get_text(strip=True)
	else:
		title = re_title.sub("", title)
		title = title.replace("&", "&amp;")
	title = brackets(title)
	title = truncate(title, 1126)
	return title  # noqa: RET504


def title_long(s: str) -> str:
	"""
	Return long title line.

	Example:
	-------
	title_long("str[ing]") -> "string".

	"""
	return s.replace("[", "").replace("]", "")


def title_short(s: str) -> str:
	"""
	Return short title line.

	Example:
	-------
	title_short("str[ing]") -> "str".

	"""
	return spaces(re_title_short.sub("", s))



# -*- coding: utf-8 -*-
# mypy: ignore-errors
#
# Copyright © 2008-2021 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

import gzip
import os
import re
from os.path import join

from pyglossary.core import log
from pyglossary.text_utils import (
	toStr,
	uintFromBytes,
)

from .bgl_reader import BGLGzipFile, BglReader, Block, FileOffS, tmpDir


def isASCII(data: str) -> bool:
	for c in data:  # noqa: SIM110
		if ord(c) >= 128:
			return False
	return True


class MetaData:
	def __init__(self) -> None:
		self.blocks = []
		self.numEntries = None
		self.numBlocks = None
		self.numFiles = None
		self.gzipStartOffset = None
		self.gzipEndOffset = None
		self.fileSize = None
		self.bglHeader = None  # data before gzip header


class MetaDataBlock:
	def __init__(self, data: bytes, _type: str) -> None:
		self.data = data
		self.type = _type


class MetaDataRange:
	def __init__(self, _type: str, count: int) -> None:
		self.type = _type
		self.count = count


class MetaData2:

	"""
	Second pass metadata.
	We need to scan all definitions in order to collect these
	statistical data.
	"""

	def __init__(self) -> None:
		# defiTrailingFields[i] - number of fields with code i found
		self.defiTrailingFields = [0] * 256
		self.isDefiASCII = True
		# isDefiASCII = true if all definitions contain only ASCII chars
		"""
		We apply a number of tests to each definition, excluding those with
		overwritten encoding (they start with <charset c=U>).
		defiProcessedCount - total number of definitions processed
		defiUtf8Count - number of definitions in utf8 encoding
		defiAsciiCount - number of definitions containing only ASCII chars
		"""
		self.defiProcessedCount = 0
		self.defiUtf8Count = 0
		self.defiAsciiCount = 0
		self.charRefs = {}  # encoding -> [ 0 ] * 257


class GzipWithCheck:

	"""
	gzip.GzipFile with check.
	It checks that unpacked data match what was packed.
	"""

	def __init__(
		self,
		fileobj,
		unpackedPath,
		reader,
		closeFileobj=False,
	) -> None:
		"""
		constructor.

		fileobj - gzip file - archive
		unpackedPath - path of a file containing original data, for testing.
		reader - reference to BglReader class instance, used for logging.
		"""
		self.file = BGLGzipFile(
			fileobj=fileobj,
			closeFileobj=closeFileobj,
		)
		self.unpackedFile = open(unpackedPath, "rb")  # noqa: SIM115
		self.reader = reader

	def __del__(self) -> None:
		self.close()

	def close(self) -> None:
		if self.file:
			self.file.close()
			self.file = None
		if self.unpackedFile:
			self.unpackedFile.close()
			self.unpackedFile = None

	def read(self, size=-1):
		buf1 = self.file.read(size)
		buf2 = self.unpackedFile.read(size)
		if buf1 != buf2:
			self.reader.msgLogFileWrite(
				f"GzipWithCheck.read: !=: size = {buf1}, ({buf2}) ({size})",
			)
		# else:
		# 	self.reader.msgLogFileWrite(
		# 		f"GzipWithCheck.read: ==: size = {buf1}, ({buf2}) ({size})",
		# 	)
		return buf1

	def seek(self, offset, whence=os.SEEK_SET):
		self.file.seek(offset, whence)
		self.unpackedFile.seek(offset, whence)
		# self.reader.msgLogFileWrite(
		# 	f"GzipWithCheck.seek: offset = {offset}, whence = {whence}",
		# )

	def tell(self):
		pos1 = self.file.tell()
		pos2 = self.unpackedFile.tell()
		if pos1 != pos2:
			self.reader.msgLogFileWrite(
				f"GzipWithCheck.tell: !=: {pos1} {pos2}",
			)
		# else:
		# 	self.reader.msgLogFileWrite(
		# 		f"GzipWithCheck.tell: ==: {pos1} {pos2}",
		# 	)
		return pos1

	def flush(self):
		if os.sep == "\\":
			pass
			# a bug in Windows
			# after file.flush, file.read returns garbage
		else:
			self.file.flush()
			self.unpackedFile.flush()


class DebugBglReader(BglReader):
	_collect_metadata2: bool = False
	_search_char_samples: bool = False
	_write_gz: bool = False
	_raw_dump_path: str = ""
	_unpacked_gzip_path: str = ""
	_char_samples_path: str = ""
	_msg_log_path: str = ""

	def open(
		self,
		filename,
	):
		if not BglReader.open(self, filename):
			return

		self.metadata2 = MetaData2() if self._collect_metadata2 else None
		if self._search_char_samples:
			self.targetCharsArray = [False] * 256
		else:
			self.targetCharsArray = None

		if self._raw_dump_path:
			self.rawDumpFile = open(self._raw_dump_path, "w", encoding="utf-8")
		if self._char_samples_path:
			self.samplesDumpFile = open(self._char_samples_path, "w", encoding="utf-8")
		if self._msg_log_path:
			self.msgLogFile = open(self._msg_log_path, "w", encoding="utf-8")

		self.charRefStatPattern = re.compile(b"(&#\\w+;)", re.IGNORECASE)

	def openGzip(self):
		with open(self._filename, "rb") as bglFile:
			if not bglFile:
				log.error(f"file pointer empty: {bglFile}")
				return False
			buf = bglFile.read(6)
			if len(buf) < 6 or buf[:4] not in {
				b"\x12\x34\x00\x01",
				b"\x12\x34\x00\x02",
			}:
				log.error(f"invalid header: {buf[:6]!r}")
				return False
			self.gzipOffset = gzipOffset = uintFromBytes(buf[4:6])
			log.debug(f"Position of gz header: {gzipOffset}")
			if gzipOffset < 6:
				log.error(f"invalid gzip header position: {gzipOffset}")
				return False

			if self._write_gz:
				self.dataFile = self._filename + "-data.gz"
				try:
					f2 = open(self.dataFile, "wb")
				except OSError:
					log.exception("error while opening gzip data file")
					self.dataFile = join(
						tmpDir,
						os.path.split(self.m_filename)[-1] + "-data.gz",
					)
					f2 = open(self.dataFile, "wb")
				bglFile.seek(gzipOffset)
				f2.write(bglFile.read())
				f2.close()
				self.file = gzip.open(self.dataFile, "rb")
				return None

			f2 = FileOffS(self._filename, gzipOffset)
			if self._unpacked_gzip_path:
				self.file = GzipWithCheck(
					f2,
					self._unpacked_gzip_path,
					self,
					closeFileobj=True,
				)
				return None

			self.file = BGLGzipFile(
				fileobj=f2,
				closeFileobj=True,
			)
			return None

	def close(self) -> None:
		BglReader.close(self)
		if self.rawDumpFile:
			self.rawDumpFile.close()
			self.rawDumpFile = None
		if self.msgLogFile:
			self.msgLogFile.close()
			self.msgLogFile = None
		if self.samplesDumpFile:
			self.samplesDumpFile.close()
			self.samplesDumpFile = None

	def __del__(self) -> None:
		BglReader.__del__(self)

	def readEntryWord(self, block, pos):
		succeed, pos, _u_word, b_word = BglReader.readEntryWord(self, block, pos)
		if not succeed:
			return
		self.rawDumpFileWriteText(f"\n\nblock type = {block.type}\nkey = ")
		self.rawDumpFileWriteData(b_word)

	def readEntryDefi(self, block, pos, b_key):
		succeed, pos, _u_defi, b_defi = BglReader.readEntryDefi(self, block, pos, b_key)
		if not succeed:
			return
		self.rawDumpFileWriteText("\ndefi = ")
		self.rawDumpFileWriteData(b_defi)

	"""
	def readEntryAlts(self, block, pos, b_key, key):
		succeed, pos, alts, b_alts = \
			BglReader.readEntryAlts(self, block, pos, b_key, key)
		if not succeed:
			return
		for b_alt in b_alts:
			self.rawDumpFileWriteText("\nalt = ")
			self.rawDumpFileWriteData(b_alt)
	"""

	def charReferencesStat(self, b_text, encoding):
		"""b_text is bytes instance."""
		# &#0147;
		# &#x010b;
		if not self.metadata2:
			return

		if encoding not in self.metadata2.charRefs:
			self.metadata2.charRefs[encoding] = [0] * 257
		charRefs = self.metadata2.charRefs[encoding]

		for index, b_part in enumerate(self.charRefStatPattern.split(b_text)):
			if index % 2 != 1:
				continue
			try:
				code = (
					int(b_part[3:-1], 16)
					if b_part[:3].lower() == "&#x"
					else int(b_part[2:-1])
				)
			except (ValueError, OverflowError):
				continue
			if code <= 0:
				continue
			code = min(code, 256)
			charRefs[code] += 1

	# write text to dump file as is
	def rawDumpFileWriteText(self, text):  # FIXME
		text = toStr(text)
		if self.rawDumpFile:
			self.rawDumpFile.write(text)

	# write data to dump file unambiguously representing control chars
	# escape "\" with "\\"
	# print control chars as "\xhh"
	def rawDumpFileWriteData(self, text):
		text = toStr(text)
		# the next function escapes too many chars, for example, it escapes äöü
		# self.rawDumpFile.write(text.encode("unicode_escape"))
		if self.rawDumpFile:
			self.rawDumpFile.write(text)

	def msgLogFileWrite(self, text):
		text = toStr(text)
		if self.msgLogFile:
			offset = self.msgLogFile.tell()
			# print offset in the log file to facilitate navigating this
			# log in hex editor
			# intended usage:
			# the log file is opened in a text editor and hex editor
			# use text editor to read error messages, use hex editor to
			# inspect char codes offsets allows to quickly jump to the right
			# place of the file hex editor
			self.msgLogFile.write(f"\noffset = {offset:#02x}\n")
			self.msgLogFile.write(text + "\n")
		else:
			log.debug(text)

	def samplesDumpFileWrite(self, text):
		text = toStr(text)
		if self.samplesDumpFile:
			offset = self.samplesDumpFile.tell()
			self.samplesDumpFile.write(f"\noffset = {offset:#02x}\n")
			self.samplesDumpFile.write(text + "\n")
		else:
			log.debug(text)

	def dumpBlocks(self, dumpPath):
		import pickle

		self.file.seek(0)
		metaData = MetaData()
		metaData.numFiles = 0
		metaData.gzipStartOffset = self.gzipOffset

		self.numEntries = 0
		self.numBlocks = 0
		range_type = None
		range_count = 0
		block = Block()
		while not self.isEndOfDictData():
			log.debug(
				f"readBlock: offset {self.file.tell():#02x}, "
				f"unpacked offset {self.file.unpackedFile.tell():#02x}",
			)
			if not self.readBlock(block):
				break
			self.numBlocks += 1

			if block.type in {1, 7, 10, 11, 13}:
				self.numEntries += 1
			elif block.type == 2:  # Embedded File (mostly Image or HTML)
				metaData.numFiles += 1

			if block.type in {1, 2, 7, 10, 11, 13}:
				if range_type == block.type:
					range_count += 1
				else:
					if range_count > 0:
						mblock = MetaDataRange(range_type, range_count)
						metaData.blocks.append(mblock)
						range_count = 0
					range_type = block.type
					range_count = 1
			else:
				if range_count > 0:
					mblock = MetaDataRange(range_type, range_count)
					metaData.blocks.append(mblock)
					range_count = 0
				mblock = MetaDataBlock(block.data, block.type)
				metaData.blocks.append(mblock)

		if range_count > 0:
			mblock = MetaDataRange(range_type, range_count)
			metaData.blocks.append(mblock)
			range_count = 0

		metaData.numEntries = self.numEntries
		metaData.numBlocks = self.numBlocks
		metaData.gzipEndOffset = self.file_bgl.tell()
		metaData.fileSize = os.path.getsize(self._filename)
		with open(self._filename, "rb") as f:
			metaData.bglHeader = f.read(self.gzipOffset)

		with open(dumpPath, "wb") as f:
			pickle.dump(metaData, f)

		self.file.seek(0)

	def dumpMetadata2(self, dumpPath):
		import pickle

		if not self.metadata2:
			return
		with open(dumpPath, "wb") as f:
			pickle.dump(self.metadata2, f)

	def processDefiStat(self, fields, defi, b_key):  # noqa: PLR0912
		BglReader.processDefiStat(self, fields, defi, b_key)

		if fields.b_title:
			self.rawDumpFileWriteText("\ndefi title: ")
			self.rawDumpFileWriteData(fields.b_title)
		if fields.b_title_trans:
			self.rawDumpFileWriteText("\ndefi title trans: ")
			self.rawDumpFileWriteData(fields.b_title_trans)
		if fields.b_transcription_50:
			self.rawDumpFileWriteText(
				f"\ndefi transcription_50 ({fields.code_transcription_50:#x}): ",
			)
			self.rawDumpFileWriteData(fields.b_transcription_50)
		if fields.b_transcription_60:
			self.rawDumpFileWriteText(
				f"\ndefi transcription_60 ({fields.code_transcription_60:#x}): ",
			)
			self.rawDumpFileWriteData(fields.b_transcription_60)
		if fields.b_field_1a:
			self.rawDumpFileWriteText("\ndefi field_1a: ")
			self.rawDumpFileWriteData(fields.b_field_1a)
		if fields.b_field_13:
			self.rawDumpFileWriteText(
				f"\ndefi field_13 bytes: {fields.b_field_13!r}",
			)
		if fields.b_field_07:
			self.rawDumpFileWriteText("\ndefi field_07: ")
			self.rawDumpFileWriteData(fields.b_field_07)
		if fields.b_field_06:
			self.rawDumpFileWriteText(
				f"\ndefi field_06: {fields.b_field_06}",
			)
		if fields.singleEncoding:
			self.findAndPrintCharSamples(
				fields.b_defi,
				f"defi, key = {b_key}",
				fields.encoding,
			)
			if self.metadata2:
				self.metadata2.defiProcessedCount += 1
				if isASCII(toStr(fields.b_defi)):
					self.metadata2.defiAsciiCount += 1
				try:
					fields.b_defi.decode("utf-8")
				except UnicodeError:
					pass
				else:
					self.metadata2.defiUtf8Count += 1
		if self.metadata2 and self.metadata2.isDefiASCII and not isASCII(fields.u_defi):
			self.metadata2.isDefiASCII = False

	# search for new chars in data
	# if new chars are found, mark them with a special sequence in the text
	# and print result into msg log
	def findAndPrintCharSamples(self, b_data: bytes, hint, encoding):
		if not self.targetCharsArray:
			return
		offsets = self.findCharSamples(b_data)
		if len(offsets) == 0:
			return
		res = ""
		utf8 = encoding.lower() == "utf-8"
		i = 0
		for o in offsets:
			j = o
			if utf8:
				while b_data[j] & 0xC0 == 0x80:
					j -= 1
			res += b_data[i:j]
			res += "!!!--+!!!"
			i = j
		res += b_data[j:]
		offsets_str = " ".join([str(el) for el in offsets])
		self.samplesDumpFileWrite(
			f"charSample({hint})\noffsets = {offsets_str}"
			f"\nmarked = {res}\norig = {b_data}\n",
		)

	def findCharSamples(self, b_data):
		"""
		Find samples of chars in b_data.

		Search for chars in data that have not been marked so far in
		the targetCharsArray array, mark new chars.
		Returns a list of offsets in b_data
		May return an empty list.
		"""
		res = []
		if not isinstance(b_data, bytes):
			log.error("findCharSamples: b_data is not a bytes instance")
			return res
		if not self.targetCharsArray:
			log.error(
				"findCharSamples: self.targetCharsArray={self.targetCharsArray}",
			)
			return res
		for i, char in enumerate(b_data):
			if char < 128:
				continue
			if not self.targetCharsArray[char]:
				self.targetCharsArray[char] = True
				res.append(i)
		return res



import logging

__all__ = [
	"MockLogHandler",
	"getMockLogger",
]


class MockLogHandler(logging.Handler):
	def __init__(self) -> None:
		logging.Handler.__init__(self)
		self.clear()

	def clear(self) -> None:
		self.recordsByLevel: dict[int, list[logging.LogRecord]] = {}

	def emit(self, record: logging.LogRecord) -> None:
		level = record.levelno
		if level in self.recordsByLevel:
			self.recordsByLevel[level].append(record)
		else:
			self.recordsByLevel[level] = [record]

	def popLog(self, level: int, msg: str, partial=False) -> "logging.LogRecord | None":
		if level not in self.recordsByLevel:
			return None
		records = self.recordsByLevel[level]
		for index, record in list(enumerate(records)):
			rec_msg = record.getMessage()
			if msg == rec_msg or (msg in rec_msg and partial):
				return records.pop(index)
		return None

	def printRemainingLogs(self, level, method: str = "") -> int:
		if level not in self.recordsByLevel:
			return 0
		count = 0
		for record in self.recordsByLevel[level]:
			count += 1
			msg = self.format(record)
			print(f"{method}: {msg!r}")
		return count

	def printRemainingErrors(self, method: str = "") -> int:
		count = self.printRemainingLogs(logging.CRITICAL, method)
		count += self.printRemainingLogs(logging.ERROR, method)
		return count

	def printRemainingwWarnings(self, method: str = "") -> int:
		return self.printRemainingLogs(logging.WARNING, method)


mockLog = None


def getMockLogger():
	global mockLog

	if mockLog is not None:
		return mockLog

	log = logging.getLogger("pyglossary")

	for handler in log.handlers:
		log.removeHandler(handler)

	mockLog = MockLogHandler()
	mockLog.setLevel(logging.WARNING)
	log.addHandler(mockLog)
	return mockLog
